#!/bin/bash
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
#
# Copyright (C) 2013, 2014 Department of Informatics and
# Telecommunications, National and Kapodistrian University of Athens
# 
#      Author: Charalampos Nikolaou <charnik@di.uoa.gr>
#
#
# This script transforms a CSP file to the undirectd graph file format of the
# MeTiS tool.
# The transformed file is written to standard output.
#
# Grap file structure:
# line 0: <nnodes> <nedges> [fmt [ncon]]
# line 1: <adjacent nodes of node 1>
# line 2: <adjacent nodes of node 2>
# line 3: <adjacent nodes of node 3>
# ...
# line <nnodes>: <adjacent nodes of node <nnodes>>
# 
# When fmt is "001", then after each adjacent node the weight of the edge
# is printed.
# 
# NOTICE:
#         * Node ids start from 0 like the CSP files.
#
# IMPORTANT:
#         * Requires GAWK 4.1 or later
#
# Run as follows:
#	$ ./csp2metis-graph.sh input.csp
#	$ cat input.csp | ./csp2metis-graph.sh
#
# Author: Charalampos (Babis) Nikolaou <charnik@di.uoa.gr>
#

awk '
BEGIN { 
	# number of nodes/edges
	nnodes = 0;
	nedges = 0;

	# adjacent matrix
	#adj

	# edge label matrix
	#lbl

	# binary encoding of RCC-8 relations
	rel["DC"]    =	  1
	rel["EC"]    =	  2
	rel["PO"]    =	  4
	rel["TPP"]   =	  8
	rel["NTPP"]  =	 16
	rel["TPPI"]  =	 32
	rel["NTPPI"] =	 64
	rel["EQ"]    =	128

	# inverse RCC-8 relations
	inv["DC"]    =   1;
	inv["EC"]    =   2;
	inv["PO"]    =   4;
	inv["TPP"]   =  32;
	inv["NTPP"]  =  64;
	inv["TPPI"]  =   8;
	inv["NTPPI"] =  16;
	inv["EQ"]    = 128;
}
{
	if (NR == 1) { # read maximum id of the nodes
		nnodes = $1 + 1;

	} else {
		if (length($0) > 1) { # not last line
			nedges++;

			# node ids in MeTiS
			s = $1;
			t = $2;

			# keep both directions by appending to the array of the
			# adjacent list for s (adj[s]) and t (adj[t]) respectively
			adj[s][length(adj[s]) + 1] = t
			adj[t][length(adj[t]) + 1] = s

			# keep also the corresponding RCC-8 relation
			r = rel[$4]
			ri = inv[$4]
			for (i = 5; i < NF; i++) {
				r = or(r, rel[$i])
				ri = or(ri, inv[$i])
			}

			lbl[s][length(lbl[s]) + 1] = r
			lbl[t][length(lbl[t]) + 1] = ri
		}
	}
}
END {
	# print number of nodes/edges/fmt
	printf "%d %d 001\n", nnodes, nedges;

	# print list of edges
	for (i = 0; i < nnodes; i++) {
		len = length(adj[i]);
		for (j = 1; j <= len; j++) {
			printf "%d %d", adj[i][j], lbl[i][j];
			if (j < len) {
				printf " ";
			}
		}
		printf "\n";
	}

}' "${1}"


