/**
 * This Source Code Form is subject to the terms of the Mozilla Public
 * License, v. 2.0. If a copy of the MPL was not distributed with this
 * file, You can obtain one at http://mozilla.org/MPL/2.0/.
 *
 * Copyright (C) 2013, 2014 Department of Informatics and
 * Telecommunications, National and Kapodistrian University of Athens
 *
 *  Created on: Dec 19, 2013
 *      Author: Charalampos Nikolaou <charnik@di.uoa.gr>
 */
#include "ConsistencyEngine.h"

ConsistencyEngine::ConsistencyEngine(PartGraph *part_graph): pgraph(part_graph) {
	csps = csp_map();
	workers = list<boost::thread*>();
	infos = list<thread_info*>();

	visited_nodes = 0;

	bytes_total = 0;
	bytes_avg = 0.0;
}

void free_csps(csp_map *csps) {
	if (csps) {
		for (csp_map::iterator it = csps->begin(); it != csps->end(); ++it) {
			FREE_PTR(it->second);
		}
		csps->clear();
	}

	FREE_PTR(csps);
}

void apply_global_split(csp_map *csps, bsplit_set *bsplit) {
	for (csp_map::iterator it = csps->begin(); it != csps->end(); ++it) {
		it->second->apply_global_split(bsplit);
	}
}


csp_map *clone_csps(csp_map *csps) {
	if (csps) {
		csp_map *clone = new csp_map();

		for (csp_map::iterator it = csps->begin(); it != csps->end(); ++it) {
			RCC8_CSP *csp = it->second->clone();

			if (csp) { // add it to the map
				(*clone)[it->first] = csp;

			} else { // make sure you release all previous clones and the map
				free_csps(clone);
				return NULL;

			}
		}

		return clone;
	}

	return NULL;
}

void ConsistencyEngine::create_csps() {
	partition_map::const_iterator itp;
	for (itp = pgraph->partitions_begin(); itp != pgraph->partitions_end(); ++itp) {
		const part_node &partition = itp->second;

		//if (!(partition.is_disconnected())) { // spawn a thread only for connected partitions
			csps[itp->first] = new RCC8_CSP(partition.get_nodes(),
											partition.get_edges(),
											pgraph->get_metis().get_xadj(),
											pgraph->get_metis().get_adjncy(),
											pgraph->get_metis().get_adjrel());
		//}

	}


	dependency_map::const_iterator ite;
	for (ite = pgraph->get_all_common_edges().cbegin(); ite != pgraph->get_all_common_edges().cend(); ++ite) {
		// for each pair of partitions, get their common edges
		const pair<idx_t, idx_t> &p = ite->first;
		const edge_list &cedges = ite->second;

		RCC8_CSP *csp_left = csps.at(p.first);
		RCC8_CSP *csp_rght = csps.at(p.second);

		csp_left->set_fixed_edges(cedges);
		csp_rght->set_fixed_edges(cedges);
	}
}

// this is completed, but inefficient
bool ConsistencyEngine::check_full_consistency_base_split() {
	long long niter = 0;
	bool consistent = true;

	std::set<idx_t> queue = std::set<idx_t>();

	// compute whether we need to split the shared edges in base relations
	bsplit_set *gl_split = compute_global_base_split(csps);

	// initialize check queue
	partition_map::const_iterator itp;
	for (itp = pgraph->partitions_begin(); itp != pgraph->partitions_end(); ++itp) {
		idx_t pid = itp->first;
		RCC8_CSP *csp = csps.at(itp->first);

		if (gl_split != NULL) {	// apply the global split set

			const split_set *lsplit = csp->get_split();
			// schedule only if it changes the network or the local network is not horn
			if (csp->apply_global_split(gl_split) || lsplit != NULL) {
				queue.insert(pid);

				if (lsplit != NULL) {
					std::cout << "local combinations: " << lsplit->get_max_value() << "\n";
				}
			}
		}
	}

	if (gl_split != NULL) {
		gl_split->incr();
		std::cout << "global combinations: " << gl_split->get_max_value() << " (#edges: " <<  gl_split->edges.size() << ")\n";
	}

	bool has_more_splits = true;
	std::set<idx_t> inconsistent = std::set<idx_t>();
	while (has_more_splits) {
		inconsistent.clear();
		niter++;
		if ((niter % 10000) == 0) {
			std::cout << "global split: checked " << niter << " splits\n";
		}

		// check consistency for CSP networks
		for (set<idx_t>::const_iterator it = queue.begin(); it != queue.end(); ++it) {
			idx_t pid = *it;

			thread_info *info = new thread_info(pid); // initialize thread information
			infos.push_back(info);

			workers.push_back(new boost::thread(consistency_command, csps.at(pid), info, true));
		}

		// waiting threads
		wait_threads(workers);
		// TODO: don't wait all threads to check for consistency. Any thread with an inconsistent
		// network should notify the main thread for such a fact
		// check consistency

		// check if any inconsistency was found
		for (list<thread_info*>::const_iterator it = infos.begin(); it != infos.end(); ++it) {
			bool local_consistent = (*it)->consistent;
			if (!local_consistent) {
				inconsistent.insert((*it)->id);
			}
		}

		// free workers
		free_threads(workers);

		// free infos
		free_infos(infos);

		// clear the queue
		queue.clear();

		if (inconsistent.empty()) { // then all CSP networks are consistent with respect to some local split
						  	  	  	// and the current global split (if any)
			has_more_splits = false;
			consistent = true;

		} else { // a CSP checked all its local split sets and found none that would make it path-consistent

//			std::cout << "which CSP checked all its options?\n";

			if (gl_split && !gl_split->end()) { // if there is another global split
//				std::cout << "increasing global split\n";

				// schedule partitions for checking
				partition_map::const_iterator itp;
				for (itp = pgraph->partitions_begin(); itp != pgraph->partitions_end(); ++itp) {
					idx_t pid = itp->first;

					RCC8_CSP *csp = csps.at(pid);

					// apply the global split set and schedule only if it changes the network
					if (csp->apply_global_split(gl_split)) {
						queue.insert(pid);

						// reset the local split iterator as well
						csp->reset_split();

					} else { // otherwise, it returned because some split was satisfiable or all
							 // were not satisfiable, OR it was already horn and it was either
							 // consistent or inconsistent
							 // in any case, since the global split does not affect it, if it was
							 // inconsistent, then the whole network is inconsistent
						if (inconsistent.find(pid) != inconsistent.end()) {
							consistent = false;
							has_more_splits = false;
							break;
						}
					}
				}

				// increment split
				gl_split->incr();

			} else {
				std::cout << "Checked all splits\n";
				has_more_splits = false;
				consistent = false;
			}
		}

	} // end of consistency checking

	FREE_BSPLIT(gl_split);

	return consistent;
}


void ConsistencyEngine::schedule_full_consistency(idx_t pid) {
	thread_info *info = new thread_info(pid); // initialize thread information
	infos.push_back(info);
	workers.push_back(new boost::thread(consistency_command, csps.at(pid), info, true));
}

void ConsistencyEngine::schedule_path_consistency_horn(idx_t pid) {
	thread_info *info = new thread_info(pid); // initialize thread information
	infos.push_back(info);
	workers.push_back(new boost::thread(consistency_command, csps.at(pid)->get_horn_csp(), info, false));
}

bool ConsistencyEngine::execute(bool pc_only) {
	if (!check_memory_requirements()) {
		cout << "Execution plan does not fit in main memory. Consider increasing the number of partitions.\n";
		return false;

	} else {
		// initialize CSP networks according to the partition graph
		create_csps();

		// check path consistency anyway
		if (!check_path_consistency(csps)) return false;

		if (pc_only || all_horn()) return true;

		// check full consistency now
		//return check_full_consistency();
		//return check_full_consistency_base_split();
		return recursive_consistency(&csps, 0);
	}
}

// Prerequisites: * the csps are path-consistent!
// 				  * the list of global constraints has been computed and ordered
// choose unprocessed edge
// if no edge is unprocessed; then return true
// compute base split for the edge
// for each base relation
//		apply the relation to all csps
//		if (full_consistent(csps)) // it first runs path-consistency and then it goes to local global consistency
//			return recursive_consistency(csps);
//		else
//			continue to the next base relation
// return false
bool ConsistencyEngine::recursive_consistency(csp_map *csps, int depth) {

	// get the next split
	bsplit_set *split = compute_next_global_constraint_to_split(*csps);

	if (split == NULL) { // no other split to make
		return full_local_consistent(csps);
	}

	assert (split->rels.size() == 1);

//	RELTYPE to_split = *(split->rels.begin());

//	cout << "next global relation (" << depth << "): " << RCC8_CSP::print_rel(to_split) << "\n";
//	cout << "depth: " << depth << "\n";

	bool local_consistent = false;
	while (!split->end() && !local_consistent) {
		visited_nodes++;
		if ((visited_nodes % 10000) == 0) {
			std::cout << "visited nodes: " << visited_nodes << "\n";
		}

//		std::cout << "\tnext base split: " << RCC8_CSP::print_rel(base_split[to_split][split->get_split_pos(0)]) << "\n";
		csp_map *new_csps = clone_csps(csps);

		if (new_csps == NULL) {
			std::cerr << "** out of memory\n";
			break;
		}

		apply_global_split(new_csps, split);

//		std::cout << "\tlevel_" << depth << ": current split: " << RCC8_CSP::print_rel(split->get_split_for_relation(0)) << "\n";
		if (check_path_consistency(*new_csps)) {
//			std::cout << "\tlevel_" << depth << ": networks are path-consistent (going deeper)\n";
			if (recursive_consistency(new_csps, depth + 1)) {
				local_consistent = true;

			} else {
//				std::cout << "\tlevel_" << depth << ": networks are INCONSISTENT ";
			}
		} else {
//			std::cout << "\tlevel_" << depth << ": networks are NOT path-consistent ";
		}

		// release resources
		free_csps(new_csps);

		split->incr();

//		if (!split->end()) {
//			std::cout << "(should try next split: " << RCC8_CSP::print_rel(split->get_split_for_relation(0)) << ")\n";
//		} else {
//			std::cout << "no other splits to check... backtracking\n";
//		}
	}

	if (!local_consistent) {
//		std::cout << "\tlevel_" << current_edge << ": checked all splits for relation (" << current_edge << "): " << RCC8_CSP::print_rel(to_split) << "\n";
	}

	// release split
	FREE_BSPLIT(split);

	return local_consistent;
}

// Prerequisite: * all global edges have been assigned a basic relation
//				 *
// We run distributed path-consistency followed by local full consistency
// If true, then we have found a tractable network for all local CSPs.
// However, some global edges might need to be split.
// If false, then some CSP checked all its splits and found none that could
// lead to a satisfying solution.
bool ConsistencyEngine::full_local_consistent(csp_map *csps) {
	// now we have refined the CSPs and we should check local full consistency
	std::set<idx_t> queue = std::set<idx_t>();

	std::cout << "executing hybrid consistency\n";

	// compute splits for all CSPs pushing to the queue any that has one
	for (csp_map::iterator it = csps->begin(); it != csps->end(); ++it) {
		if (it->second->compute_split_set()) {
			queue.insert(it->first);
		}
	}

	assert (queue.empty() == false);

	/* list of worker threads on csps */
	list<boost::thread*> workers = list<boost::thread*>();

	/* list of information on thread execution */
	list<thread_info*> infos = list<thread_info*>();

	bool consistent = true;
//	while (queue.size() > 0 && consistent) {
		// check consistency for CSP networks
		for (set<idx_t>::const_iterator it = queue.begin(); it != queue.end(); ++it) {
			idx_t pid = *it;

			thread_info *info = new thread_info(pid); // initialize thread information
			infos.push_back(info);

			workers.push_back(new boost::thread(consistency_command, csps->at(pid), info, true));
		}

		// waiting threads
		wait_threads(workers);

		// check if any inconsistency was found
		consistent = is_consistent(infos);

		// free workers
		free_threads(workers);

		// free infos
		free_infos(infos);

		queue.clear();

/*
		if (consistent) {
			// check whether common subnetworks between partitions are different;
			// in case they are, refine them and add the refined partitions to be checked again
			csp_map tocheck;
			for (csp_map::iterator it = csps->begin(); it != csps->end(); ++it) {
				RCC8_CSP *csp = it->second->get_horn_csp();
				if (csp == NULL) { // this CSP has not splits, but we should not overwrite its network, because we might backtrack later on
					csp = it->second->clone();
				}

				tocheck[it->first] = it->second;
			}

			consistent &= check_path_consistency(tocheck);

			if (!consistent) {
				// then, some edge of the horn CSPs OR the cloned ones is zero
				// if there are CSPs that have other splits to check, then we should schedule

				// TODO remember to free up memory from the above cloning
			}
		} // otherwise, some CSP checked all local splits and found none that is satisfiable
	}
*/
	return consistent;
}

bsplit_set* ConsistencyEngine::compute_next_global_constraint_to_split(const csp_map &csps) {
	// compute the global split
	bsplit_set *gl_split = compute_global_base_split(csps);
	bsplit_set *toreturn = NULL;

//	static bool first = true;
	if (gl_split) { // if there exists a global split
		//if (first) {
//			std::cout << "global combinations: " << gl_split->get_max_value() << " (edges: " << gl_split->rels.size() << ")\n";
//			std::cout << "global edges: " << gl_split->rels.size() << "\n";
//			first = false;
//		}

		list<rcc8_con> global_cons;
		edge_list::const_iterator itedges = gl_split->edges.begin();
		std::list<RELTYPE>::const_iterator itrels = gl_split->rels.begin();

		// add all constraints that are to be split in a list
		for (; itrels != gl_split->rels.end(); ++itrels, ++itedges) {
			global_cons.push_back(boost::make_tuple(itedges->first, itedges->second, *itrels));
		}

		// sort them according to their restrictiveness (first would be the more restrictive one)
		global_cons.sort(compare_constraints_vb);

		// we do not need the global split any more
		FREE_PTR(gl_split);

		if (!global_cons.empty()) {
			idx_t src = global_cons.begin()->get<0>();
			idx_t trgt = global_cons.begin()->get<1>();
			RELTYPE tosplit = global_cons.begin()->get<2>();

			toreturn = create_atomic_bsplit_set(std::make_pair(src, trgt), tosplit);
		}
	}

	return toreturn;
}

bool ConsistencyEngine::check_path_consistency(const csp_map &tocheck) {
		/* partitions that need to be checked again for consistency after they have been refined */
		set<int> need_check = set<int>();

		/* list of worker threads on csps */
		list<boost::thread*> workers = list<boost::thread*>();

		/* list of information on thread execution */
		list<thread_info*> infos = list<thread_info*>();

		/* number of iterations (i.e., refinements) */
		int niter = -1;

		bool consistent = true;

		// insert all networks to be checked
		for (csp_map::const_iterator it = tocheck.begin(); it != tocheck.end(); ++it) {
			need_check.insert(it->first);
		}

		while (need_check.size() > 0 && consistent) {
			niter++;

			// run path-consistency for each network
			set<int>::iterator its;
			for (its = need_check.begin(); its != need_check.end(); ++its) {
				thread_info *info = new thread_info(*its); // initialize thread information
				infos.push_back(info);

				workers.push_back(new boost::thread(consistency_command, tocheck.at(*its), info, false));
			}

			// waiting threads
			wait_threads(workers);

			// TODO: don't wait all threads to check for consistency. Any thread with an inconsistent
			// network should notify the main thread for such a fact
			// check consistency
			consistent &= is_consistent(infos);

			// free workers
			free_threads(workers);

			// free infos
			free_infos(infos);

			// empty queue
			need_check.clear();

			if (consistent) {
				// check whether common subnetworks between partitions are different;
				// in case they are, refine them and add the refined partitions to be checked again
				const edge_list& part_edges = pgraph->get_edges();
				edge_list::const_iterator it;
				for (it = part_edges.begin(); it != part_edges.end(); ++it) {
					const edge_list &shared_edges = pgraph->get_common_edges(it->first, it->second);

					// debug
					//std::cout << "checking partitions " << it->first << ", " << it->second << "\n";

					int	res = RCC8_CSP::check_and_refine(tocheck.at(it->first), tocheck.at(it->second), shared_edges);

					if (res == INCONSISTENT) { // inconsistent
						consistent = false;
						break;

					} else if (res == REFINED_LEFT) {
						need_check.insert(it->first);

					} else if (res == REFINED_RIGHT) {
						need_check.insert(it->second);

					} else if (res == REFINED_BOTH) {
						need_check.insert(it->first);
						need_check.insert(it->second);

					} else {
						assert (res == UNCHANGED);
					}
				}

			} else {
//				std::cout << "found local inconsistency while executing path-consistency\n";
			}
		} // end of path-consistency

		return consistent;
}

//void ConsistencyEngine::check_and_refine(bool horn, std::set<idx_t> &refine, std::set<idx_t> &inconsistent) {
//	std::list<std::pair<idx_t, idx_t> > problematic_pairs;
//
//	// check whether common subnetworks between partitions are different;
//	// in case they are, refine them and add the refined partitions to be checked again
//	const list<pair<idx_t, idx_t> >& part_edges = pgraph->get_edges();
//	list<pair<idx_t, idx_t> >::const_iterator it;
//	for (it = part_edges.begin(); it != part_edges.end(); ++it) {
//		const list<pair<idx_t, idx_t> > &shared_edges = pgraph->get_common_edges(it->first, it->second);
//
//		RCC8_CSP *left =  (horn ? csps.at(it->first)->get_horn_csp():csps.at(it->first));
//		RCC8_CSP *right = (horn ? csps.at(it->second)->get_horn_csp():csps.at(it->second));
//
//		int res = RCC8_CSP::check_and_refine(left, right, shared_edges);
//
//		if (res == INCONSISTENT) { // inconsistent
//			inconsistent.insert(it->first);
//			inconsistent.insert(it->second);
//			problematic_pairs.push_back(*it);
//
//		} else if (res == REFINED_LEFT) {
//			refine.insert(it->first);
//
//		} else if (res == REFINED_RIGHT) {
//			refine.insert(it->second);
//
//		} else if (res == REFINED_BOTH) {
//			refine.insert(it->first);
//			refine.insert(it->second);
//
//		} else {
//			assert (res == UNCHANGED);
//		}
//	}
//
//	list<pair<idx_t, idx_t> >::const_iterator itl;
//	for (itl = problematic_pairs.begin(); itl != problematic_pairs.end(); ++itl) {
//		refine.erase(itl->first);
//		refine.erase(itl->second);
//	}
//}

bool ConsistencyEngine::is_consistent(const list<thread_info*> &infos) {
	bool consistent = true;
	// gather and combine individual consistency results
	for (list<thread_info*>::const_iterator it = infos.begin(); it != infos.end(); ++it) {
			consistent &= (*it)->consistent;
	}

	return consistent;
}

bool ConsistencyEngine::check_memory_requirements() const {
//	cout << "Memory requirements: " << Util::get_human_size(bytes_total) << endl;
//	cout << "Average memory requir. (per partition): " << Util::get_human_size(bytes_avg) << endl;
//	cout << "System memory: " << Util::get_human_size(Util::get_system_memory()) << endl;

	return (pgraph->get_memory_requirements() < Util::get_system_memory()) ? true:false;
}

ConsistencyEngine::~ConsistencyEngine() {
	for (csp_map::iterator it = csps.begin(); it != csps.end(); ++it) {
		FREE_PTR(it->second);
	}
	csps.clear();
}

void consistency_command(RCC8_CSP *csp, thread_info *info, bool full_consistency) {
	if (full_consistency) { // run full consistency
		info->consistent = csp->consistent();

	} else { // run path-consistency
		info->consistent = csp->path_consistent();
		//std::cout << info->id << ": path consistency finished with result = " << info->consistent << "\n";
	}
}

bsplit_set* ConsistencyEngine::compute_global_base_split(const csp_map &csps) const {
	std::set<edge_t> split_edges;
	bsplit_set *bsplit = NULL;

	static bool first = true;
	int num_base = 0;

	dependency_map::const_iterator ite;
	for (ite = pgraph->get_all_common_edges().cbegin(); ite != pgraph->get_all_common_edges().cend(); ++ite) {
		// for each pair of partitions, get their common edges

		edge_list::const_iterator itel;
		for (itel = ite->second.begin(); itel != ite->second.end(); ++itel) { // for each common edge

			// try to put it in the set of edges that are going to participate in a split
			std::pair<std::set<edge_t>::iterator, bool> ins_res = split_edges.insert(*itel);
			if (ins_res.second) { // if it was inserted right now
				RCC8_CSP *holding_left = csps.at((ite->first).first);
				RCC8_CSP *holding_right = csps.at((ite->first).second);

				RELTYPE rel_left = holding_left->get_relation(itel->first, itel->second);
				RELTYPE rel_right = holding_right->get_relation(itel->first, itel->second);

				if (num_base_rels(rel_left) == 1) {
					num_base++;
				}

				assert (rel_left == rel_right);

				bsplit = create_or_update_base_split(bsplit, rel_left, *itel);
			}
		}
	}

	// complete initialization of base split set
	if (bsplit != NULL) {
		bsplit->it = new bit_set(3 * bsplit->rels.size());
		//std::cout << "global edges: " << bsplit->rels.size() << endl;
	}

	if (first && num_base) {
		std::cout << "PC made " << num_base << " relations to be base\n";
		first = false;
	}

	return bsplit;
}

bool ConsistencyEngine::sanity_check(bool consistent) const {
	MetisGraph &mgraph = pgraph->get_metis();
	idx_t size = 2 * mgraph.num_edges();
	idx_t diffs = 0;
	std::stringstream diffs_str;

	RELTYPE *rels = new RELTYPE[size];
	std::fill_n(rels, size, -1);

	RELTYPE *part = new RELTYPE[size];

	for (csp_map::const_iterator it = csps.begin(); it != csps.end(); ++it) {
		const RCC8_CSP *csp = it->second;

		for (int i = 0; i < mgraph.num_nodes(); i++) {
			for (int j = mgraph.xadj[i]; j < mgraph.xadj[i + 1]; j++) {
				if (i < mgraph.adjncy[j]) {
					idx_t reals = i;
					idx_t realt = mgraph.adjncy[j];

					nodeid_map::const_iterator itn = csp->nodeids.find(reals);

					if (itn != csp->nodeids.end()) {
						idx_t source = itn->second;

						itn = csp->nodeids.find(realt);
						if (itn != csp->nodeids.end()) {
							idx_t target = itn->second;

							ENSURE_ORDER(source, target);

							RELTYPE rel = csp->csp[TRINDX(source, target)];

							if (rels[j] == -1 || rel == 0) { // mark inconsistency anyway
								rels[j] = rel;
								part[j] = it->first;

							} else if (rels[j] != rel) {
								diffs++;

								diffs_str << "\t";
								diffs_str << RCC8_CSP::print_rel(rels[j]);
								diffs_str << " != ";
								diffs_str << RCC8_CSP::print_rel(rel);
								diffs_str << ": (" << reals << ", " << realt << ") ";
								diffs_str << "[" << part[j] << "-" << it->first << "]\n";
							}
						}
					}
				}
			}
		}
	}

	// check if any edge is zero
	idx_t zero = 0;
	idx_t init = 0;
	for (idx_t i = 0; i < size; i++) {
		if (rels[i] == 0) {
			zero++;

		} else if (rels[i] == -1) {
			init++;
		}
	}
	init -= size / 2;

	std::cout << "\nSanity check report\n";
	std::cout << "-------------------\n";

	if (init) {
		std::cout << " * There was/were " << init << " edge(s) that was/were not touched by any CSP (THIS IS A BUG)\n";
	}

	if (zero) {
		std::cout << " * There was/were " << zero << " zero edge(s) in the graph (COMPARE THIS WITH THE CONSISTENCY RESULT)\n";
	}

	if (consistent && diffs > 0) {
		std::cout << " * Some CSPs disagree on common edges (THIS IS A BUG)" << "\n";
		std::cout << " * Total differences are " << diffs << "/" << size << "\n";
		std::cout << diffs_str.str();
	}

	if (!zero && !init && (diffs == 0 || !consistent)) {
		std::cout << " * All OK\n";
	}

	delete rels;
	delete part;

	return diffs == 0;
}

void ConsistencyEngine::update_metis_graph() {
	MetisGraph &mgraph = pgraph->get_metis();

	for (csp_map::const_iterator it = csps.begin(); it != csps.end(); ++it) {
		const RCC8_CSP *csp = it->second;

		for (int i = 0; i < mgraph.num_nodes(); i++) {
			for (int j = mgraph.xadj[i]; j < mgraph.xadj[i + 1]; j++) {
				if (i < mgraph.adjncy[j]) {
					idx_t reals = i;
					idx_t realt = mgraph.adjncy[j];

					nodeid_map::const_iterator itn = csp->nodeids.find(reals);

					if (itn != csp->nodeids.end()) {
						idx_t source = itn->second;

						itn = csp->nodeids.find(realt);
						if (itn != csp->nodeids.end()) {
							idx_t target = itn->second;

							ENSURE_ORDER(source, target);

							// update relation and weight
							mgraph.adjrel[j] = csp->csp[TRINDX(source, target)];
							mgraph.adjwgt[j] = num_base_rels(mgraph.adjrel[j]) * compute_weight_vb(mgraph.adjrel[j]);
						}
					}
				}
			}
		}
	}
}


