/**
 * This Source Code Form is subject to the terms of the Mozilla Public
 * License, v. 2.0. If a copy of the MPL was not distributed with this
 * file, You can obtain one at http://mozilla.org/MPL/2.0/.
 *
 * Copyright (C) 2013, 2014 Department of Informatics and
 * Telecommunications, National and Kapodistrian University of Athens
 *
 *  Created on: Dec 19, 2013
 *      Author: Charalampos Nikolaou <charnik@di.uoa.gr>
 */
#ifndef CONSISTENCYENGINE_H_
#define CONSISTENCYENGINE_H_

#include <boost/thread/thread.hpp>

#include "PartGraph.h"
#include "rcc8csp.h"

#include <algorithm>
#include <list>

#define CAN_SPLIT(s)	((s) != NULL && (s)->has_next())

typedef struct thread_info {
	int id;				/**< the id of the thread corresponding to the partition number */
	bool consistent;	/**< true when the checked RCC-8 network was consistent */

	explicit thread_info(int identifier):id(identifier) { consistent = false;}
} thread_info;

typedef boost::unordered_map<idx_t, RCC8_CSP*> csp_map;
typedef boost::unordered_map<idx_t, thread_info*> thread_info_map;
//typedef boost::unordered_map<idx_t, boost::thread*> thread_map;

void consistency_command(RCC8_CSP *csp, thread_info *info, bool global_phase);

class ConsistencyEngine {
private:
	PartGraph *pgraph;				/**< partition graph over which the execution of the
										 consistency algorithm will take place */

	csp_map csps;					/**< RCC-8 networks for checking consistency */
	list<boost::thread*> workers;	/**< list of worker threads on \a csps */
	list<thread_info*> infos;		/**< list of information on thread execution */

	long visited_nodes;

	long int bytes_total;			/**< memory requirements in bytes */
	float bytes_avg;				/**< average memory requirements in bytes */

	bool check_memory_requirements() const;
	bool check_path_consistency(const csp_map &tocheck);
	void create_csps();

	void schedule_full_consistency(idx_t pid);
	void schedule_path_consistency_horn(idx_t pid);

	void check_and_refine(bool horn, std::set<idx_t> &need_check, std::set<idx_t> &inconsistent);

	bsplit_set* compute_next_global_constraint_to_split(const csp_map &csps);
	bool check_full_consistency_base_split();

	bsplit_set* compute_global_base_split(const csp_map &) const;

	inline bool all_horn() {
		for (csp_map::iterator it = csps.begin(); it != csps.end(); ++it) {
			if (!(it->second->is_horn())) {
				return false;
			}
		}

		return true;
	}

	bool recursive_consistency(csp_map *csps, int depth);


	bool full_local_consistent(csp_map *csps);

	inline static void wait_threads(list<boost::thread*> &workers) {
		for (list<boost::thread*>::iterator it = workers.begin(); it != workers.end(); ++it) {
			(*it)->join();

		} // all threads stopped
	}

	inline static void free_threads(list<boost::thread*> &workers) {
		for (list<boost::thread*>::iterator it = workers.begin(); it != workers.end(); ++it) {
			delete *it;
		}
		workers.clear();
	}

	inline static void free_infos(list<thread_info*> &infos) {
		for (list<thread_info*>::iterator it = infos.begin(); it != infos.end(); ++it) {
			delete *it;
		}
		infos.clear();
	}

	static bool is_consistent(const list<thread_info*> &infos);
public:
	ConsistencyEngine(PartGraph *part_grpah);

	bool execute(bool pc_only = false);
	void update_metis_graph();
	bool sanity_check(bool consistent) const;
	virtual ~ConsistencyEngine();
};

#endif /* CONSISTENCYENGINE_H_ */
