/**
 * This Source Code Form is subject to the terms of the Mozilla Public
 * License, v. 2.0. If a copy of the MPL was not distributed with this
 * file, You can obtain one at http://mozilla.org/MPL/2.0/.
 *
 * Copyright (C) 2013, 2014 Department of Informatics and
 * Telecommunications, National and Kapodistrian University of Athens
 *
 *  Created on: Dec 17, 2013
 *      Author: Charalampos Nikolaou <charnik@di.uoa.gr>
 */
#include "MetisGraph.h"

MetisGraph::MetisGraph(const char *file) {
	filename = file;
	nparts = 0;
	ncon = 1;
	nvtxs = 0;
	xadj = NULL;
	adjncy = NULL;
	adjwgt = NULL;
	adjrel = NULL;
	nedges = 0;

	objval = 0;
	parts = NULL;

	// statistics
	parts_size = NULL;
	smallest_idx = 0;
	biggest_idx = 0;
	zero_sz_parts = 0;
	r_stdev = 0.0;
	avg_degree = 0.0;

	// set defaults
	set_defaults();
}

//void MetisGraph::recompute_weights() {
//	idx_t sz = 2 * nedges;
//	for (idx_t i = 0; i < sz; i++) {
//		adjwgt[i] = num_base_rels(adjrel[i]) * compute_weight_vb(adjrel[i]);
//	}
//}

bool MetisGraph::partition(idx_t npartitions) {
	if (nparts != 0) { // repartition
		// initialize partitioning-dependent values
		std::fill_n(parts, nvtxs, -1);

		FREE_ARRAY(parts_size);

		graphs.clear();

		objval = 0;

		smallest_idx = 0;
		biggest_idx = 0;
	}

	nparts = npartitions;

	// do the actual partitioning
	if (!partition_internal()) {
		nparts = 0;
		return false;
	}

	// initialize graphs/fringes for partitions
	graphs = std::vector<edge_set>(nparts);
	fringes = std::vector<std::list<std::pair<idx_t, idx_t> > >(nparts);

	// initialize size table for partitions
	parts_size = new idx_t[nparts];
	std::fill_n(parts_size, nparts, 0);

	// compute partitions' sizes, graphs, and fringes
	for (int i = 0; i < nvtxs; i++) {
		parts_size[parts[i]] += 1;

		// get neighbors
		for (int j = xadj[i]; j < xadj[i + 1]; j++) {
			// if they belong to the same partition
			if (parts[i] == parts[adjncy[j]]) {
				if (i < adjncy[j]) { // keep it only once
					graphs[parts[i]].insert(make_pair(i, adjncy[j]));
				}

			} else { // otherwise, update fringe information
				fringes[parts[i]].push_back(make_pair(i, parts[adjncy[j]]));
			}
		}
	}

	float avg_load = (float) nvtxs / float (nparts);
	// calculate partition statistics
	for (int i = 0; i < nparts; i++) {
		if (parts_size[i] == 0) {
			zero_sz_parts++;

			if (i == smallest_idx) {
				smallest_idx++;
			}
		} else {
			if (parts_size[i] < parts_size[smallest_idx]) {
				smallest_idx = i;
			}

			if (parts_size[i] > parts_size[biggest_idx]) {
				biggest_idx = i;
			}
		}

		// calculate relative standard deviation
		r_stdev += powf((parts_size[i] - avg_load), 2.0);
	}

	r_stdev = sqrtf(r_stdev / nparts) / 100;

	return true;
}

bool MetisGraph::partition_internal() {
	int metis_code;
	if ((metis_code = METIS_PartGraphKway(&nvtxs, &ncon, xadj, adjncy,
										NULL, NULL, adjwgt,
										&nparts,
										NULL, NULL,
										options,
										&objval, parts)) != METIS_OK) {

		if (metis_code == METIS_ERROR_INPUT) {
			cerr << "MeTiS: " << METIS_INPUT_ERROR;

		} else if (metis_code == METIS_ERROR_MEMORY) {
			cerr << "MeTiS: " << METIS_MEM_ERROR;

		} else if (metis_code == METIS_ERROR) {
			cerr << "MeTiS: " << METIS_INT_ERROR;
		}

		return false;
	}

	return true;

}

edge_set& MetisGraph::edgecut(edge_set &edgecut, edgecut_map &edgecut_edges) const {

	for (int i = 0; i < nvtxs; i++) {
		for (int j = xadj[i]; j < xadj[i + 1]; j++) {
			if (parts[i] != parts[adjncy[j]]) {
				if (parts[i] < parts[adjncy[j]]) { // put it only once
					edgecut.insert(make_pair(parts[i], parts[adjncy[j]]));

					// insert edges for this edgecut
					edgecut_map::iterator it = edgecut_edges.find(make_pair(parts[i], parts[adjncy[j]]));
					if (it == edgecut_edges.end()) { // edgecut not found, insert it
						pair<edgecut_map::iterator, bool> res = edgecut_edges.insert(edgecut_map::value_type(make_pair(parts[i], parts[adjncy[j]]), list<pair<idx_t, idx_t> >()));

						if (i < adjncy[j]) { // put it always ordered
							(res.first)->second.push_back(make_pair(i, adjncy[j]));

						} else {
							(res.first)->second.push_back(make_pair(adjncy[j], i));
						}

					} else { // edgecut found
						if (i < adjncy[j]) { // put it always ordered
							it->second.push_back(make_pair(i, adjncy[j]));

						} else {
							it->second.push_back(make_pair(adjncy[j], i));
						}
					}
				}
			}
		}
	}

	return edgecut;
}

set<idx_t>& MetisGraph::fringe_for_edgecut(idx_t spart, idx_t tpart, set<idx_t>& vertices) const {

	list<pair<idx_t, idx_t> >::const_iterator it;
	for (it = fringes[spart].begin(); it != fringes[spart].end(); ++it) {
		if (it->second == tpart) {
			vertices.insert(it->first);
		}
	}

	return vertices;
}

void MetisGraph::augment_fringe_for_edgecut(idx_t spart, idx_t via, const edge_set &edgecuts,
																	const set<idx_t>& vertices,
																	edge_set &updated_edgecuts) {
	std::set<idx_t> s_tparts;
	std::set<idx_t> v_tparts;

	// find the partitions connected with partition spart and via
	edge_set::const_iterator it;
	for (it = edgecuts.begin(); it != edgecuts.end(); ++it) {
		// spart
		if (it->first == spart && it->second != via) {
			s_tparts.insert(it->second);

		} else if (it->second == spart && it->first != via) {
			s_tparts.insert(it->first);
		}

		// via
		if (it->first == via && it->second != spart) {
			v_tparts.insert(it->second);

		} else if (it->second == via && it->first != spart) {
			v_tparts.insert(it->first);
		}
	}

	// refine the list of partitions taking the intersection
	std::set<idx_t> tparts;
	std::set_intersection(s_tparts.begin(), s_tparts.end(),
						  v_tparts.begin(), v_tparts.end(),
						  std::inserter(tparts, tparts.begin()));

	list<pair<idx_t, idx_t> >::const_iterator itl;
	for (itl = fringes[via].begin(); itl != fringes[via].end(); ++itl) { // for each pair (node, part)
																		 // in the fringe of via
		if (tparts.find(itl->second) != tparts.end() &&
			vertices.find(itl->first) != vertices.end()) { // if part == tpart and node \in vertices
			// put node into the fringe list of partition spart associating to the edgecut for tpart
			fringes[spart].push_back(std::make_pair(itl->first, itl->second));

			// mark the updated edgecuts (always in order)
			if (spart < itl->second) {
				updated_edgecuts.insert(std::make_pair(spart, itl->second));

			} else {
				updated_edgecuts.insert(std::make_pair(itl->second, spart));
			}
		}
	}

//	std::cout << "s_tparts size: " << s_tparts.size() << "\n";
//	std::cout << "v_tparts size: " << v_tparts.size() << "\n";
//	std::cout << "size of intersection: " << tparts.size() << "\n\n";

}

edge_list & MetisGraph::induced_graph(idx_t part, const set<idx_t> &vertices, edge_list & indg) const {
	if (vertices.size() > 1) {
		// you can optimize it further by checking the maximum number of
		// edges for the specified set (n * (n-1) / 2)
		// iterate over all edges of the partition
//		idx_t edges_seen = 0;
//		idx_t max_sz = vertices.size();
//		max_sz *= (max_sz - 1) / 2 + 1;
//		&& edges_seen <= max_sz
		edge_set::const_iterator it;
		for (it = graphs[part].begin(); it != graphs[part].end(); ++it) {
			// check if both endpoints are present in the requested vertices list
			if (vertices.find(it->first) != vertices.end() &&
				vertices.find(it->second) != vertices.end()) {
				indg.push_back(make_pair(it->first, it->second));
//				edges_seen++;
			}
		}
	}

	return indg;
}

void MetisGraph::print_stats() const {
	std::cout << "MeTiS statistics\n";
	std::cout << "----------------\n";
	std::cout << "Total edgecut         : " << objval << endl;
	std::cout << "Average partition size: " << ceil((float) nvtxs/ (float) nparts) << "\n";
	std::cout << "Smaller partition size: " << parts_size[smallest_idx] << "\n";
	std::cout << "Biggest partition size: " << parts_size[biggest_idx] << "\n";
	std::cout << "Zero size partitions  : " << zero_sz_parts << "\n";
	std::cout << "Relative standard dev.: " << r_stdev << "\n";
	std::cout << "Average degree        : " << avg_degree << "\n";

	if (nparts <= 16) {
		for (int i = 0; i < nvtxs; i++) {
			//cout << "Partition #" << i << ": " << parts_size[i];

			if (nvtxs <= 50) {
				cout << " [ ";
				for (int j = 0; j < nvtxs; j++) {
					if (parts[j] == i) {
						cout << j;
						edge_list::const_iterator it;
						for (it = fringes[i].begin(); it != fringes[i].end(); ++it) {
							if (it->first == j) {
								cout << "*"; // is in the fringe set
								break;
							}
						}
						cout << " ";
					}
				}
				cout << "]";
			}

			//cout << endl;
		}
	}

	// debug (edgecut)
//	for (int i = 0; i < nvtxs; i++) {
//		for (int j = xadj[i]; j < xadj[i + 1]; j++) {
//			if (parts[i] != parts[adjncy[j]] && i < adjncy[j]) {
//				cout << "edgecut: (" << i << ", " << adjncy[j] << ")";
//				cout << "[" << parts[i] << "-" << parts[adjncy[j]] << "]" << endl;
//			}
//		}
//	}
}

bool MetisGraph::read(bool peek) {
	if (xadj == NULL) { // read input graph only once
		if (!MetisGraph::read_metis_graph(filename, &nvtxs, &nedges, &xadj, &adjncy, &adjwgt, &adjrel, peek)) {
			return false;
		}

		if (!peek) {
			parts = new idx_t[nvtxs];	/* returned partitioning */

			// compute average degree
			for (int i = 0; i < nvtxs; i++) {
				avg_degree += xadj[i + 1] - xadj[i];
				//std::cout << "*" << xadj[i + 1] - xadj[i] << "\n";
			}

			avg_degree /= nvtxs;
		}
	}

	return true;
}

/**
 * Reads a MeTiS graph.
 */
bool MetisGraph::read_metis_graph(const char *filename, idx_t *nvtxs, idx_t *nedges,
								  idx_t **xadj, idx_t **adjncy,
								  idx_t **adjwgt, RELTYPE **adjrel, bool peek) {
	ifstream graph(filename);

	if (!graph.is_open()) {
		cerr << "MeTiS: error while opening file " << filename << endl;
		return false;
	}

	string line;
	bool header_seen = false;

	int adjidx = 0;	 	/* adjacent index */
	idx_t nidx = 0; 	/* current node idx */
	while (!graph.eof()) {
		getline(graph, line);

		if (graph.eof()) {
			break;
		}

		if (line[0] == '%') { // comment
			continue; // skip
		}

		if (!header_seen) { // parse header
			stringstream strm(line);

			// read number of vertices and number of edges
			strm >> *nvtxs;
			strm >> *nedges;
			// TODO read fmt parameter (currently ignored and supposed to be always equal to 001)

			if (peek) { // skip reading the file
				return true;
			}

			*xadj = new idx_t[*nvtxs + 1];
			*adjncy = new idx_t[2 * *nedges];
			*adjwgt = new idx_t[2 * *nedges];
			*adjrel = new RELTYPE[2 * *nedges];

			header_seen = true;

			// initialize
			(*xadj)[0] = 0;

			continue;
		}

		// read adjacent list
		stringstream strm(line);

		while (!strm.eof()) {
			strm >> (*adjncy)[adjidx];
			strm >> (*adjrel)[adjidx];

			// compute weight
			adjwgt[0][adjidx] = num_base_rels((*adjrel)[adjidx]) * compute_weight_vb((*adjrel)[adjidx]);
			//std::cout << adjwgt[0][adjidx] << " " << RCC8_CSP::print_rel((*adjrel)[adjidx]) << "\n";

			adjidx++;
		}

		(*xadj)[nidx + 1] = adjidx;

		nidx++;
	}

	//debug
	/*
	cout << "nodes: " << *nvtxs << endl;
	cout << "edges: " << nedges << endl;
	for (int i = 0; i < *nvtxs; i++) {
		cout << "Adjacent list for node " << i << ":";
		for (int j = (*xadj)[i]; j < (*xadj)[i + 1]; j++) {
			cout << " " << (*adjncy)[j];
		}
		cout << endl;
	}

	cout << "  xadj: ";
	for (int i = 0; i < *nvtxs + 1; i++) {
		cout << (*xadj)[i] << " ";
	}
	cout << endl;

	cout << "adjncy: ";
	for (int i = 0; i < 2 * *nedges; i++) {
		cout << (*adjncy)[i] << " ";
	}
	cout << endl;
	*/

	graph.close();

	return true;
}
