/**
 * This Source Code Form is subject to the terms of the Mozilla Public
 * License, v. 2.0. If a copy of the MPL was not distributed with this
 * file, You can obtain one at http://mozilla.org/MPL/2.0/.
 *
 * Copyright (C) 2013, 2014 Department of Informatics and
 * Telecommunications, National and Kapodistrian University of Athens
 *
 *  Created on: Dec 17, 2013
 *      Author: Charalampos Nikolaou <charnik@di.uoa.gr>
 */
#ifndef METISGRAPH_H_
#define METISGRAPH_H_

#include <iostream>
#include <fstream>
#include <sstream>

#include <utility>
#include <vector>

#include <cmath>

#include <algorithm>

using namespace std;

#include "metis/metis.h"

#include "datastruct.h"
#include "Util.h"
#include "rcc8.h"
#include "rcc8csp.h"

// MeTiS errors
#define METIS_INT_ERROR		"MeTiS internal error";
#define METIS_INPUT_ERROR 	"MeTiS input error";
#define METIS_MEM_ERROR		"MeTiS memory allocation error";

//
#define OBJ_MINCUT	"C"
#define OBJ_VOLUME	"V"

#define WEIGHTED_PARTITION true

class MetisGraph {
	friend class PartGraph; // this should be removed finally
	friend class ConsistencyEngine;

private:
	const char *filename; 	/**< the filename of the metis graph */
	idx_t nparts; 			/**< number of partitions */
	idx_t ncon;				/**< number of balancing constraints (default is 1) */

	idx_t nvtxs;			/**< number of vertices in the input graph */
	idx_t nedges;			/**< number of edges in the input graph */
	idx_t *xadj;			/**< the adjacent index matrix of the CSR graph representation */
	idx_t *adjncy;			/**< the adjacent matrix of the CSR graph representation */
	idx_t *adjwgt;			/**< the weights of the edges of \a adjncy indexed by \a xadj */
	RELTYPE *adjrel;		/**< the RCC-8 relations of the edges of \a adjncy indexed by \a xadj */

	idx_t objval;			/**< returned objective value (from optimization) */
	idx_t *parts;			/**< returned partitioning */

	// statistics
	idx_t *parts_size; 		/**< total number of nodes per partition */
	idx_t smallest_idx; 	/**< smallest partition index */
	idx_t biggest_idx;		/**< biggest partition index */
	int zero_sz_parts;		/**< number of partitions with zero nodes */
	float r_stdev;			/**< relative standard deviation of the partitioning (partition balance) */
	float avg_degree;		/**< average degree number */

	vector<edge_set> graphs;					/**< graphs per partition */
	vector<edge_list> fringes;					/**< set of (node, part) in the fringe of a partition */

	idx_t options[METIS_NOPTIONS]; /**< options for partitioning */

	bool partition_internal();

	inline void cleanup() {
		FREE_ARRAY(xadj);
		FREE_ARRAY(adjncy);
		FREE_ARRAY(adjwgt);
		FREE_ARRAY(adjrel);
		FREE_ARRAY(parts);
		FREE_ARRAY(parts_size);

		graphs.clear();
		fringes.clear();
	}

	inline void set_defaults() {
		METIS_SetDefaultOptions(options);

		// set specific options
		options[METIS_OPTION_PTYPE] = METIS_PTYPE_KWAY;
		options[METIS_OPTION_OBJTYPE] = METIS_OBJTYPE_CUT; //METIS_OBJTYPE_VOL | METIS_OBJTYPE_CUT
		options[METIS_OPTION_NUMBERING] = 0; // starts vertex numbering from index 0
		options[METIS_OPTION_MINCONN] = 1;
		//options[METIS_OPTION_CONTIG] = 1; // forces contiguous partitioning
		//options[METIS_OPTION_NO2HOP] = 1;
	}

	//void recompute_weights();

public:
	MetisGraph(const char *filename);

	/**
	 * Return the number of nodes of the graph.
	 *
	 * @return
	 */
	inline idx_t num_nodes() const {
		return nvtxs;
	}

	/**
	 * Return the number of edges of the graph.
	 *
	 * @return
	 */
	inline idx_t num_edges() const {
		return nedges;
	}

	/**
	 * Returns the number of computed partitions.
	 *
	 * @return
	 */
	inline idx_t num_parts() const {
		return nparts;
	}

	inline const idx_t* get_xadj() const {
		return xadj;
	}

	inline const idx_t* get_adjncy() const {
		return adjncy;
	}

	inline const RELTYPE* get_adjrel() const {
		return adjrel;
	}

	inline float get_average_degree() const {
		return avg_degree;
	}

	/**
	 * Sets the objective for partitioning to be the cut-size.
	 * This is the default behavior.
	 */
	inline void set_objective_mincut() {
		options[METIS_OPTION_OBJTYPE] = METIS_OBJTYPE_CUT;
	}

	/**
	 * Sets the objective for partitioning to be the total communication
	 * volume between the partitions.
	 */
	inline void set_objective_volume() {
		options[METIS_OPTION_OBJTYPE] = METIS_OBJTYPE_VOL;
	}

	/**
	 * Returns the name/alias/descriptor string of the objective that is
	 * currently used for partitioning.
	 *
	 * @return one of \a OBJ_MINCUT or \a OBJ_VOLUME (see the respective #define)
	 */
	inline const char *get_obj_name() const {
		return (options[METIS_OPTION_OBJTYPE] == METIS_OBJTYPE_CUT) ? OBJ_MINCUT:OBJ_VOLUME;
	}

	inline std::set<idx_t> get_graph_nodes () const {
		std::set<idx_t> gnodes = std::set<idx_t>();

		for (int i = 0; i < nvtxs; i++) {
			gnodes.insert(i);
		}

		return gnodes;
	}

	inline edge_set get_graph_edges() const {
		edge_set edgelist = edge_set();

		for (int i = 0; i < nvtxs; i++) {
			for (int j = xadj[i]; j < xadj[i + 1]; j++) {
				if (i < adjncy[j]) {
					edgelist.insert(std::make_pair(i, adjncy[j]));
				}
			}
		}

		return edgelist;
	}

	/**
	 * Returns the weight (RCC-8 label) of the given edge.
	 *
	 * @param edge the edge for which we search its label
	 * @return the RCC-8 relation of the given edge specified in \a edge
	 */
	inline RELTYPE get_weight(const edge_t &edge) const {
		return adjrel[Util::csr_edge_index(edge.first, edge.second, xadj, adjncy)];
	}

	/**
	 * Return the number of non-empty partitions.
	 *
	 * @return
	 */
	inline idx_t num_nonempty_parts() const {
		int non_empty = 0;
		for (int i = 0; i < nparts; i++) {
			if (parts_size[i] > 0) {
				non_empty++;
			}
		}

		return non_empty;
	}

	inline idx_t part_size(idx_t id) const {
		return parts_size[id];
	}

	/**
	 * @return the relative standard deviation
	 */
	inline float get_r_stdev() const {
		return r_stdev;
	}

	/**
	 * Computes a set of paired partitions participating to an edgecut and
	 * associates it with the corresponding set of edges.
	 *
	 * Two partitions participate to an edgecut if one or more nodes of the
	 * one partition are connected to one or more nodes of the other. Such
	 * nodes are in the fringe set of each partition.
	 *
	 * @param edgecut a set that will be filled with pairs of partitions
	 * @param edgecut_edges a map associating an edgecut to a set of edges
	 * @return a reference to the edgecut set
	 */
	edge_set& edgecut(edge_set &edgecut, edgecut_map &edgecut_edges) const;

	/**
	 * Computes the set of vertices of the fringe of partition \a spart that
	 * are connected with the fringe of partition \a tpart.
	 *
	 * @param spart partition id
	 * @param tpart partition id
	 * @param vertices the set of vertices to populate
	 * @return a subset of the fringe of partition \a spart
	 */
	set<idx_t>& fringe_for_edgecut(idx_t spart, idx_t tpart, set<idx_t>& vertices) const;

	/**
	 * spart ---- via ---- tpart
	 *   |___________________|
	 *
	 * If there exists a partition \a tpart such that the partitions \a spart and \a via
	 * are connected to \a tpart through an edgecut, then augment the fringe of
	 * partition \a spart with the subset of the specified set of vertices that are in
	 * the fringe of partition \a via with respect to an edgecut betweet \a via and \a tpart.
	 *
	 * Prerequisite: an edgecut should already exist between \a spart and \a via
	 *
	 * @param spart the partition the fringe of which will be updated
	 * @param via the partition holding the vertices \a vertices
	 * @param edgecuts the set of edgecuts of the underlying graph
	 * @param vertices a set of vertices of partition \a via
	 * @param updated_edgecuts [out]
	 */
	void augment_fringe_for_edgecut(idx_t spart, idx_t via, const edge_set &edgecuts,
															const set<idx_t>& vertices,
															edge_set &updated_edgecuts);

	/**
	 * Returns the graph for the specified partition. If the graph for the
	 * partition contains disconnected nodes, then these nodes are not included
	 * in the returned graph.
	 *
	 * @param part the id of the partition
	 * @return list of paired vertices, i.e., list of edges
	 */
	inline const edge_set &graph(idx_t part) const {
		return graphs[part];
	}

	inline void augment_graph(idx_t part, const edge_list &edges) {
		//std::cout << "before: " << graphs[part].size() << "\n";
		graphs[part].insert(edges.begin(), edges.end());
		//std::cout << "after: " << graphs[part].size() << "\n";
	}

	/**
	 * Computes the induced graph for a subset of the vertices of a partition.
	 *
	 * @param part the id of the partition.
	 * @param vertices the set of vertices
	 * @param indg where to place the induced graph, @see return value
	 * @return the induced graph of the partition for the specified subset of vertices
	 */
	edge_list & induced_graph(idx_t part, const set<idx_t> &vertices, edge_list &indg) const;

	/**
	 * Partitions the underlying graph into the given number of partitions.
	 *
	 * Prerequisites: The underlying data structures should have been initialized
	 * by calling the method \a read().
	 *
	 * @param nparts [in] the number of partitions to create.
	 * @return false/true whether the partition was successful or not.
	 */
	bool partition(idx_t nparts);

	inline idx_t total_edgecut() const {
		return objval;
	}

	void print_stats() const;

	inline ~MetisGraph() {
		cleanup();
	}

	/**
	 * Parses the input graph and populates the underlying data structures.
	 *
	 * @param peek [in,optional] if true it just reads the number of nodes and edges
	 * @return true for success, false for failure
	 */
	bool read(bool peek = false);

	/**
	 * Parses a file in the Metis graph file format and constructs the graph representation
	 * in the CSR format.
	 *
	 * @param filename [in] the filename of the graph file.
	 * @param nvtxs [out] the number of the nodes the graph contains.
	 * @param nedges [out] the number of the edges the graph contains
	 * @param xadj [out] the index of the adjacent array (see next parameter \a adjncy).
	 * @param adjncy [out] an array holding the adjacent nodes indexed by array \a xadj.
	 * @param adjwgt [out] an array holding the weights of edges indexed by array \a xadj.
	 * @param peek [in,optional] if true it just reads the number of nodes and edges
	 * @return
	 */
	static bool read_metis_graph(const char *filename, idx_t *nvtxs, idx_t *nedges, idx_t **xadj,
													   idx_t **adjncy, idx_t **adjwgt, RELTYPE **adjrel,
													   bool peek = false);
};



#endif /* METISGRAPH_H_ */
