/**
 * This Source Code Form is subject to the terms of the Mozilla Public
 * License, v. 2.0. If a copy of the MPL was not distributed with this
 * file, You can obtain one at http://mozilla.org/MPL/2.0/.
 *
 * Copyright (C) 2013, 2014 Department of Informatics and
 * Telecommunications, National and Kapodistrian University of Athens
 *
 *  Created on: Dec 17, 2013
 *      Author: Charalampos Nikolaou <charnik@di.uoa.gr>
 */
#include "PartGraph.h"

PartGraph::PartGraph(MetisGraph &metis, std::string descr, int parts, const char* strategy): mgraph(metis), desc(descr) {
	partitions = partition_map();
	init_parts = parts;
	heuristic = strategy;

	nvirtual = 0;
	disconnected_parts = 0;

	// statistics
	big_part = -1;
	small_part = -1;
	big_sz = std::numeric_limits<int>::min();
	small_sz = std::numeric_limits<int>::max();

	big_virt_part = -1;
	small_virt_part = -1;
	big_virt_sz = std::numeric_limits<int>::min();
	small_virt_sz = std::numeric_limits<int>::max();

	big_node_inc = std::numeric_limits<float>::min();
	small_node_inc = std::numeric_limits<float>::max();
	total_node_inc = 0.0;

	r_stdev = 0.0;
	memory = 0;

	need_more_parts = false;

	total_time = 0;

	base_combs = 0;
	num_shared_edges = 0;
}

bool PartGraph::add_or_update_edge(idx_t srcp, idx_t trgtp, const edge_list &common_edges) {
	edge_t edge = make_pair(srcp, trgtp);

	dependency_map::iterator dmi = shared_edges.find(edge);
	if (dmi == shared_edges.end()) { // edge dependency not found
		edges.push_back(edge);
		shared_edges[make_pair(srcp, trgtp)] = common_edges;

		return false;

	} else { // found, thus update putting them at the back
		dmi->second.insert(dmi->second.end(), common_edges.begin(), common_edges.end());

		return true;
	}


}

/**
 * @deprecated
 * @param edgecuts
 */
void PartGraph::update_edges(const edge_set &edgecuts) {
	edge_set::const_iterator it;
	for (it = edgecuts.begin(); it != edgecuts.end(); ++it) {
		// consolidate partitions by filling missing edges
		part_node &one = get_partition(it->first);
		part_node &two = get_partition(it->second);

		consolidate_partitions(one, two, *it, false);
		consolidate_partitions(two, one, *it, false);
	}
}

// this is a safe way to fill missing edges between two nodes that are present
// in two partitions that have common edges.
// any time that has been run with min_inter, it hasn't added any further edge
// in the partitioning graph (thus it does not change it at all). Therefore, we can
// iterate over the edges of the partitioning graph and not over all pairs of partitions
// In general, we should not count on this method when creating the Partitioning Graph.
// But if it is run and fill missing edges, that means that our algorithm for partitioning
// is INCOMPLETE!
void PartGraph::finalize(const edge_set &cut) {
//	std::cout << "\n*** Sanity checking of the partitioning graph ***\n";
//	std::cout << "--------------------------------------------------\n";
	bool missing_edge_added = false;

	edge_set::iterator its;
	for (its = cut.begin(); its != cut.end(); ++its) {
		part_node &one = get_partition(its->first);
		part_node &two = get_partition(its->second);

		//if (!(one.receives_from(two))) {
			missing_edge_added |= consolidate_partitions(one, two, *its, false);

		//} else {
			missing_edge_added |= consolidate_partitions(two, one, *its, false);
		//}

	}

//	partition_map::const_iterator it1, it2;
//	for (it1 = partitions.begin(); it1 != partitions.end(); ++it1) {
//		for (it2 = partitions.begin(); it2 != partitions.end(); ++it2) {
//			if (it1->second.id() < it2->second.id()) {
//				part_node &one = get_partition(it1->second.id());
//				part_node &two = get_partition(it2->second.id());
//
//				missing_edge_added |= consolidate_partitions(one, two, std::make_pair(it1->second.id(), it2->second.id()), true);
//				missing_edge_added |= consolidate_partitions(two, one, std::make_pair(it1->second.id(), it2->second.id()), true);
//			}
//		}
//	}

//	if (missing_edge_added) {
//		std::cout << "--------------------------------------------------\n";
//		std::cout << " * Found missing edges and they were filled (THIS IS A BUG)\n";
//		std::cout << " * Check that there are no missing edges above marked with * (THAT WOULD BE ANOTHER BUG)\n";
//
//	} else {
//		std::cout << " * All OK\n";
//	}
}

bool PartGraph::consolidate_partitions(const part_node &one, part_node &two, const edge_t &p, bool debug) {
	const edge_set &one_edges = one.get_edges();

	const std::set<idx_t> &two_nodes = two.get_nodes();
	edge_set &two_edges = two.get_edges();

	bool missing_edge = false;
	edge_set::const_iterator it;
	for (it = one_edges.begin(); it != one_edges.end(); ++it) {
		if (two_nodes.find(it->first)  != two_nodes.end() &&
			two_nodes.find(it->second) != two_nodes.end()) {

			assert (it->first < it->second);

			if (two_edges.insert(*it).second) {
				missing_edge = true;

				// update shared edges or create one
				bool created = !add_or_update_edge(p.first, p.second, edge_list(1, *it));

				if (debug) {//debug
					std::cout << "\tinserted missing edge (" << it->first << ", " << it->second << ") ";
					std::cout << "[" << p.first << "-" << p.second << "]";
					if (created) {
						std::cout << "*";
					}
					std::cout << "\n";
				}
			}
		}
	}

	return missing_edge;
}

split_set* PartGraph::compute_global_split() const {
	std::set<edge_t> split_edges;
	split_set *split = NULL;
	long three_splits = 0; 				/* number of edges with three splits */

	dependency_map::const_iterator ite;
	for (ite = shared_edges.begin(); ite != shared_edges.end(); ++ite) {
		// for each pair of partitions, get their common edges

		edge_list::const_iterator itel;
		for (itel = ite->second.begin(); itel != ite->second.end(); ++itel) { // for each common edge

			// try to put it in the set of edges that are going to participate in a split
			std::pair<std::set<edge_t>::iterator, bool> ins_res = split_edges.insert(*itel);
			if (ins_res.second) { // if it was inserted right now
				RELTYPE rel = mgraph.get_weight(*itel);
				bool three_split = false;	/* true when the inserted edge can be split in three horn relations */
				split = create_or_update_split(split, rel, *itel, &three_split);
				if (three_split) {
					three_splits++;
				}
			}
		}
	}

	// complete initialization of split set
	if (split != NULL) {
		split->it = new bit_set(split->edges.size());
		split->msb_it = new bit_set(three_splits);
	}

	return split;
}

void PartGraph::compute_base_split_combinations() {
	edge_set split_edges;
	long double combs = 1;

	dependency_map::const_iterator ite;
	for (ite = shared_edges.begin(); ite != shared_edges.end(); ++ite) {
		// for each pair of partitions, get their common edges

		edge_list::const_iterator itel;
		for (itel = ite->second.begin(); itel != ite->second.end(); ++itel) { // for each common edge

			// try to put it in the set of edges that are going to participate in a split
			std::pair<edge_set::iterator, bool> ins_res = split_edges.insert(*itel);
			if (ins_res.second) { // if it was inserted right now
				RELTYPE rel = mgraph.get_weight(*itel);
				combs *= get_num_bsplits(rel);
			}
		}
	}

	base_combs = combs;
	num_shared_edges = split_edges.size();
}

cost_type PartGraph::get_cost() const {
	float cost = 0.0;
	float WEIGHT_EDGE = 0.7;
	float WEIGHT_PART_SIZE = 0.1;
	float WEIGH_LOAD_BALANCE = 0.2;
	//float WEIGHT_INTER = 0.4;

	float clique_edge_number = partitions.size() * (partitions.size() - 1) / 2;
	cost += WEIGHT_EDGE * ((float) edges.size() / clique_edge_number);
	//get_memory_requirements() Util::get_system_memory();

	//cost += WEIGHT_EDGE * edges.size();

	//cost += WEIGHT_PART_SIZE * (1.0 / total_node_inc);
	cost += WEIGHT_PART_SIZE * ((float) memory / (float) Util::get_system_memory());

//	dependency_map::const_iterator itm;
//	int inter = 0;
//	for (itm = shared_edges.begin(); itm != shared_edges.end(); ++itm) {
//		inter += itm->second.size() + 1;
//	}
//
//	// percentage of the memory spent to intersections over the total memory needs
//	cost += WEIGHT_INTER * ((float) memory_shared / (float) memory);

	// load balance of partitions
	cost += WEIGH_LOAD_BALANCE * r_stdev;

	//cout << desc << "	" << cost << endl;
	//return cost;
	return base_combs;
	//return num_shared_edges;
}

void PartGraph::calc_stats() {
	int total_nodes = 0;
	int connected_parts = 0;
	int total_nodes_in_connected_parts = 0;

	// calculate smallest/biggest partition and max/min percentage in node increase
	partition_map::const_iterator itp;
	for (itp = partitions.begin(); itp != partitions.end(); ++itp) {
		const part_node &partition = itp->second;

		int init_sz = partition.init_size();
		int sz = partition.size();
		float incr = 0;

		if (partition.is_disconnected()) {
			disconnected_parts++;

		} else { // increase memory requirements
			memory += RCC8_CSP::memory_needs(sz);

			connected_parts++;
			total_nodes_in_connected_parts += sz;

			if (sz > PART_THRESHOLD_LOAD) {
				need_more_parts = true;
			}
		}

		if (init_sz > 0) {
			incr = (float) sz / (float) init_sz;
		}

		// update total number of nodes
		total_nodes += sz;

		if (!partition.is_virtual()) {
			total_node_inc += init_sz;

			if (incr > big_node_inc) {
				big_node_inc = incr;
			}

			if (incr < small_node_inc) {
				small_node_inc = incr;
			}

			if (partition.size() < small_sz) {
				small_part = partition.id();
				small_sz = partition.size();
			}

			if (partition.size() > big_sz) {
				big_part = partition.id();
				big_sz = partition.size();
			}

		} else { // virtual case
			if (partition.size() < small_virt_sz) {
					small_virt_part = partition.id();
					small_virt_sz = partition.size();
			}

			if (partition.size() > big_virt_sz) {
				big_virt_part = partition.id();
				big_virt_sz = partition.size();
			}
		}
	}

	if (partitions.size() > 0) {
		total_node_inc = (float) total_nodes / total_node_inc;

		float avg_load = total_nodes_in_connected_parts / connected_parts;

		// calculate relative standard deviation
		partition_map::const_iterator itp;
		for (itp = partitions.begin(); itp != partitions.end(); ++itp) {
			const part_node &partition = itp->second;
			r_stdev += powf((partition.size() - avg_load), 2.0);
		}

		r_stdev = sqrtf(r_stdev / connected_parts) / 100;
	}

	compute_base_split_combinations();
}

void PartGraph::print_stats() const {

	std::cout << "Partitioning graph statistics for **" << desc << "**\n";
	std::cout << "----------------------------------------------------\n";
	std::cout << "Total cost              : " << get_cost() << "\n";
	std::cout << "Number of edges         : " << edges.size() << "\n";
	std::cout << "Number of parts         : " << partitions.size() << "\n";
	std::cout << "Number of discon. parts : " << disconnected_parts << "\n";
	std::cout << "Number of virtual parts : " << nvirtual << "\n";
	std::cout << "Number of base combs.   : " << base_combs << " (edges: " << num_shared_edges << ")" << "\n";
	std::cout << "Biggest partition size  : " << big_sz << " (id: " << big_part << ")" << "\n";
	std::cout << "Smallest partition size : " << small_sz << " (id: " << small_part << ")" << "\n";

	if (nvirtual > 0) {
		std::cout << "Biggest virtual partition size : " << big_virt_sz << " (id: " << big_virt_part << ")" << "\n";
		std::cout << "Smallest virtual partition size: " << small_virt_sz << " (id: " << small_virt_part << ")" << "\n";
	}

	std::cout << "Maximum node increase   : " << big_node_inc << "x" << "\n";
	std::cout << "Minimum node increase   : " << small_node_inc << "x" << "\n";
	std::cout << "Total node increase     : " << total_node_inc << "x" << "\n";
	std::cout << "Relative standard dev.  : " << r_stdev << "\n";
	std::cout << "Memory requirements     : " << Util::get_human_size(memory) << "\n";
	if (need_more_parts) {
		std::cout << "*** Please consider increasing the number of partitions ***\n";
	}

	//debug
//	list<pair<idx_t, idx_t> >::const_iterator it;
//	for (it = edges.begin(); it != edges.end(); ++it) {
//		cout << "(" << it->first << ", " << it->second << ")\n";
//	}
//
//	for (unsigned int i = 0; i < partitions.size(); i++) {
//		cout << "Graph for partition #" << i << ": ";
//		partitions[i].print();
//		cout << endl;
//	}

	std::cout << "\n";
}

PartGraph::~PartGraph() {
}



