/**
 * This Source Code Form is subject to the terms of the Mozilla Public
 * License, v. 2.0. If a copy of the MPL was not distributed with this
 * file, You can obtain one at http://mozilla.org/MPL/2.0/.
 *
 * Copyright (C) 2013, 2014 Department of Informatics and
 * Telecommunications, National and Kapodistrian University of Athens
 *
 *  Created on: Dec 17, 2013
 *      Author: Charalampos Nikolaou <charnik@di.uoa.gr>
 */
#ifndef PARTGRAPH_H_
#define PARTGRAPH_H_

#include "MetisGraph.h"
#include "part_node.h"
#include "rcc8csp.h"

#define PART_THRESHOLD_LOAD 8000

typedef long double cost_type;

class PartGraph {
private:
	MetisGraph &mgraph;					/**< the MeTiS graph object for partitioning the input */

	std::string desc;					/**< a textual description of the parameters used to create this
	 	 	 	 	 	 	 	 	 	 	 partitioning graph */
	const char *heuristic;				/**< the name of the partitioning heuristic that will be used */
	partition_map partitions;			/**< partition nodes */
	edge_list edges;					/**< edges between partition nodes */
	dependency_map shared_edges;		/**< shared edges between two partitions connected with an edge */

	int init_parts;						/**< initial number of partitions */
	int nvirtual;						/**< number of virtual partitions */
	int disconnected_parts;				/**< number of connected partitions */

	// statistics
	int big_part;						/**< biggest partition */
	int small_part;						/**< smallest partition */
	int big_sz;							/**< biggest partition size */
	int small_sz;						/**< smallest partition size */

	int big_virt_part;					/**< biggest virtual partition */
	int small_virt_part;				/**< smallest virtual partition */
	int big_virt_sz;					/**< biggest virtual partition size */
	int small_virt_sz;					/**< smallest virtual partition size */

	float small_node_inc;				/**< lowest increase in nodes among all partitions */
	float big_node_inc;					/**< greatest increase in nodes among all partitions */
	float total_node_inc;				/**< total increase in node size across partitions */

	float r_stdev;						/**< relative standard deviation (load balancing of partitions) */
	long memory;						/**< memory requirements */

	bool need_more_parts;				/**< true when we need to partition in more parts (see the
											 PART_THRESHOLD_LOAD constant) */

	long total_time;					/**< total time for constructing this partition graph */

	long double base_combs;				/**< the number of base combinations for the shared edges */
	int num_shared_edges;				/**< number of shared edges */

public:
	PartGraph(MetisGraph &, std::string descr, int parts, const char *strategy);
	~PartGraph();

	const char* get_name() const {
		return heuristic;
	}

	cost_type get_cost() const;

	inline void set_total_time(long duration) {
		total_time = duration;
	}

	inline long get_total_time() const {
		return total_time;
	}

	inline bool need_more_partitions() const {
		return need_more_parts;
	}

	split_set* compute_global_split() const;

	/**
	 * Returns the memory requirements for this partitioning.
	 *
	 * @return the memory requirements in bytes
	 */
	inline long get_memory_requirements() const {
		return memory;
	}

	void compute_base_split_combinations();

	inline int size() const {
		return partitions.size();
	}

	inline const std::string& get_desc() const {
		return desc;
	}

	/**
	 * Get the underlying MeTiS object.
	 *
	 * @return a reference to the MeTiS object
	 */
	inline MetisGraph& get_metis() {
		return mgraph;
	}

	/**
	 * Get the underlying MeTiS object.
	 *
	 * @return a const reference to the MeTiS object
	 */
	inline const MetisGraph& get_metis() const {
		return mgraph;
	}

	/**
	 * Creates a new partition node in the partition graph with
	 * the given graph (set of edges).
	 *
	 * Multiple
	 * calls of this function with the same identifier do not
	 * add the node multiple times or re-initialize any of their
	 * underlying data members.
	 *
	 * @param id the identifier of the partition
	 */
	inline void create_partition(idx_t id, const edge_set &graph) {
		partitions.insert(partition_map::value_type(id, part_node(id, false, graph)));
	}

	/**
	 * Returns a modifiable reference to a certain partition.
	 *
	 * @param id the identifier of the partition to retrieve
	 * @return a modifiable reference to the partition
	 */
	inline part_node& get_partition(idx_t id) {
		return partitions.at(id);
	}

	/**
	 * Returns a const reference to a certain partition.
	 *
	 * @param id the identifier of the partition to retrieve
	 * @return a modifiable reference to the partition
	 */
	inline const part_node& get_partition(idx_t id) const {
		return partitions.at(id);
	}

	inline const partition_map::const_iterator partitions_begin() const {
		return partitions.cbegin();
	}

	inline const partition_map::const_iterator partitions_end() const {
		return partitions.cend();
	}

	/**
	 * Adds a new edge between two partitions in the partitioning graph associating
	 * to it a list of edges. If this edge is already existing, then it updates the
	 * associated list of edges by augmenting it with the given list.
	 *
	 * @param srcp id of the source partition
	 * @param trgtp id ot the target partition
	 * @param common_edges list of edges to associate with the edge (\a srcp, \a trgtp)
	 * @return false if the edge was created by this call, true if existed before
	 */
	bool add_or_update_edge(idx_t srcp, idx_t trgtp, const edge_list &common_edges);

	void update_edges(const edge_set &edgecuts);

	void finalize(const edge_set &cut);
	bool consolidate_partitions(const part_node &one, part_node &two, const edge_t &p, bool debug);

	inline const edge_list& get_common_edges(idx_t srcp, idx_t trgtp) const {
		return shared_edges.at(make_pair(srcp, trgtp));
	}

	inline const dependency_map& get_all_common_edges() const {
		return shared_edges;
	}

	inline const edge_list & get_edges() const {
		return edges;
	}

	void calc_stats();

	/**
	 * Prints various statistical information computed for the
	 * internals of this class.
	 *
	 * Prerequisite: method \a calc_stats() should have been called prior to this call.
	 */
	void print_stats() const;

};
#endif /* PARTGRAPH_H_ */


