/**
 * This Source Code Form is subject to the terms of the Mozilla Public
 * License, v. 2.0. If a copy of the MPL was not distributed with this
 * file, You can obtain one at http://mozilla.org/MPL/2.0/.
 *
 * Copyright (C) 2013, 2014 Department of Informatics and
 * Telecommunications, National and Kapodistrian University of Athens
 *
 *  Created on: Dec 19, 2013
 *      Author: Charalampos Nikolaou <charnik@di.uoa.gr>
 */
#include "PartGraphPicker.h"

PartGraphPicker::PartGraphPicker(const char *filename): mgraph(MetisGraph(filename)) {
	init();
}

std::string PartGraphPicker::create_summary_desc(int nparts, int parts_requested, p_strat pstrategy, const char *obj) {
	std::stringstream strm;

	strm << "P" << nparts; 							// number of partitions
	if (nparts != parts_requested) {
		strm << "/" << parts_requested;
	}

	strm << "-"; 									// delimiter
	strm << "S_" << strategy_alias(pstrategy);		// partition strategy
	strm << "-"; 									// delimiter
	strm << "OBJ_" << obj;							// partitioning objective

	return strm.str();
}

bool PartGraphPicker::partition(int parts, p_strat strategy, bool volume) {
	long s_time = util::get_time();

	// read input graph first
	bool res = mgraph.read();
	parse_time = util::get_time() - s_time;

	if (!res) {
		return false;
	}

	// set objective function
	if (volume) mgraph.set_objective_volume();

	if (mgraph.partition(parts)) {
		//mgraph.print_stats();
		create_partition_graph(parts, strategy);
	}

	return true;
}

PartGraph* PartGraphPicker::create_partition_graph(int nparts, p_strat strategy) {
	std::string desc = create_summary_desc(mgraph.num_nonempty_parts(), nparts, strategy, mgraph.get_obj_name());
	PartGraph *pgraph = new PartGraph(mgraph, desc, mgraph.num_nonempty_parts(), strategy_alias(strategy));
	pgraphs.push_back(pgraph);

	//std::cout << desc;
	/* create partition graph */
	(*p_strat_impl[strategy]) (mgraph, pgraph);
	//std::cout << ".\n";

	return pgraph;
}

bool PartGraphPicker::compute_partitions() {
	long s_time = util::get_time();

	// read input graph first
	bool res = mgraph.read();
	parse_time = util::get_time() - s_time;

	if (!res) {
		return false;
	}

	// compute partitions based on a number of options
	int i = 1, p = 2;
	for (; p <= 1024; i++, p <<= 1) { // for partitions 2 to 1024, increasing by 2
		if ((float) mgraph.num_nodes() / (float) p > 1.0) { // check whether it is reasonable to partition in p parts

			int n_part_opts = 2;	/* number of alternatives for partitioning */
			while (n_part_opts--) {
				if (n_part_opts == 1) {
					mgraph.set_objective_mincut();

				} else {
					mgraph.set_objective_volume();
				}

				if (mgraph.partition(p)) {
					if (mgraph.num_nonempty_parts() == 1) {
						std::cout << "partitioning strategy gave 1 part; skipping\n";
						continue;
					}

					for (p_strat s = PSTRAT_BEGIN; s != PSTRAT_END; ++s) { // partition using all strategies
						if (s == BLINKERED) continue;
						/* create partition graph */
						create_partition_graph(p, s);
					}
				}
			}
		}
	}

	return true;
}

PartGraphPicker::~PartGraphPicker() {
	FREE_VEC(pgraphs);
	pgraphs.clear();
}

