/**
 * This Source Code Form is subject to the terms of the Mozilla Public
 * License, v. 2.0. If a copy of the MPL was not distributed with this
 * file, You can obtain one at http://mozilla.org/MPL/2.0/.
 *
 * Copyright (C) 2013, 2014 Department of Informatics and
 * Telecommunications, National and Kapodistrian University of Athens
 *
 *  Created on: Dec 19, 2013
 *      Author: Charalampos Nikolaou <charnik@di.uoa.gr>
 */
#ifndef PARTGRAPHPICKER_H_
#define PARTGRAPHPICKER_H_

#include <vector>
#include <limits>

#include "MetisGraph.h"
#include "PartGraph.h"
#include "p_strat.h"
#include "rcc8.h"

/**
 * Compare two partitioning graphs based on their cost. If the costs are equal,
 * then the one with the lower difference between the requested number of partitions
 * and the number of non-empty partitions is preferred.
 *
 * @param p1 a pointer to a partitioning graph
 * @param p2 a pointer to a partitioning graph
 * @return true if \a p1 comes before \a p2 in an ordering
 */
inline bool compare_PartGraph(const PartGraph *p1, const PartGraph *p2) {
	cost_type cost1 = p1->get_cost();
	cost_type cost2 = p2->get_cost();

	return (cost1 == cost2) ? (p1->get_metis().num_parts() - p1->get_metis().num_nonempty_parts() < p2->get_metis().num_parts() - p2->get_metis().num_nonempty_parts()):(cost1 < cost2);
}

class PartGraphPicker {
private:
	MetisGraph mgraph;					/**< the MeTiS graph object for partitioning the input */

	std::vector<PartGraph*>	pgraphs;	/**< partition graphs based on different partition strategies */

	// statistics
	long int parse_time;				/**< time to read the input graph file (by MeTiS) */
	long int partition_time;			/**< time to compute possible partitionings */

	/**
	 * Creates a textual representation that is easily readable and parsable for keeping
	 * track of the way (parameters) a partitioning graph is created.
	 *
	 * @param nparts [in] the actual number of partitions
	 * @param parts_requested [in] the requested number of partitions
	 * @param pstrategy [in] the partition strategy
	 * @param obj[in] the objective for partitioning
	 * @return a summary text description easily readable/parsable for a partitioning graph
	 */
	std::string create_summary_desc(int nparts, int parts_requested, p_strat pstrategy, const char *obj);

	/**
	 * Creates a partitioning graph using a specific partitioning strategy. Partitions
	 * that are empty by MeTiS are not included in the partition graph. Hence, the
	 * partitioning graph may contain fewer number of partitions than the requested
	 * one.
	 *
	 * Prerequisites: The underlying graph should have been already read
	 * and partitioned by MeTiS.
	 *
	 * @param parts the number of partitions
	 * @param strategy the strategy to use for partitioning
	 * @return a pointer to the created PartGraph object represented the
	 * executed partitioning
	 */
	PartGraph* create_partition_graph(int nparts, p_strat strategy);

	inline void init() {
		// initialize statistics variables
		parse_time = 0;
		partition_time = 0;
	}

public:
	PartGraphPicker(const char *filename);

	void initialize() {
		FREE_VEC(pgraphs);
		pgraphs.clear();
	}

	/**
	 * Computes a number of different partitionings by trying a number of combinations
	 * between a) the number of partitions, b) options passed to MeTiS, and c) the
	 * partition strategy. This method reads also the input graph to partition from
	 * a file.
	 *
	 * @return false in case the input graph file is not found, true otherwise
	 */
	bool compute_partitions();

	/**
	 * Partitions a graph using a specific partitioning strategy. This method
	 * reads also the input graph to partition from a file.
	 *
	 * @param parts [in] the number of partitions
	 * @param strategy [in] the strategy to use for partitioning
	 * @param volume [in] if true, sets the objective function to be the total
	 * 		  communication volume among partitions
	 * @return false in case the input graph file is not found, true otherwise
	 */
	bool partition(int parts, p_strat strategy, bool volume);

	inline PartGraph* pick_best_partition() {
		std::sort(pgraphs.begin(), pgraphs.end(), compare_PartGraph);
		return pgraphs[0];
	}

	virtual ~PartGraphPicker();
};

#endif /* PARTGRAPHPICKER_H_ */
