/**
 * This Source Code Form is subject to the terms of the Mozilla Public
 * License, v. 2.0. If a copy of the MPL was not distributed with this
 * file, You can obtain one at http://mozilla.org/MPL/2.0/.
 *
 * Copyright (C) 2013, 2014 Department of Informatics and
 * Telecommunications, National and Kapodistrian University of Athens
 *
 *  Created on: Dec 19, 2013
 *      Author: Charalampos Nikolaou <charnik@di.uoa.gr>
 */
#ifndef UTIL_H_
#define UTIL_H_

#include <sstream>
#include <sys/time.h>

#include "metis/metis.h"

/* macros for quickly freeing memory allocated by the *new* operator */
#define FREE_ARRAY(array)	if ((array) != NULL) { delete [] (array); (array) = NULL; }
#define FREE_PTR(pointer)	if ((pointer) != NULL) { delete (pointer); (pointer) = NULL; }
#define FREE_VEC(vec)		for (unsigned int _i_ = 0; _i_ < (vec).size(); _i_++) { FREE_PTR((vec)[_i_]); }

#if defined __APPLE__ && __MACH__ || macintosh || Macintosh

#define APPLE_OS
#include <sys/types.h>
#include <sys/sysctl.h>

#elif defined __linux__

#define LINUX_OS
#include <sys/unistd.h>

#else

// fallback
#define UNIX_OS
#include <sys/unistd.h>

#endif

namespace util {
	inline long int get_time() {
        struct timeval timestr;

        gettimeofday(&timestr, NULL);
        return timestr.tv_usec + 1000000 * timestr.tv_sec;
	}

	inline std::string print_time(long int duration) {
		std::stringstream strm;
		strm << duration / 1000 << " ms ("<< duration / 1000000 <<" sec)";

		return strm.str();
	}

	/**
	 * Returns the given duration in seconds.
	 *
	 * @param duration
	 * @return
	 */
	inline double get_time(long int duration) {
		return (double) duration / 1000000.0;
	}
}

class Util {
public:
	Util();

	/**
	 * Gets the index of the edge in the \a adjncy (or \a adjwgt) matrix.
	 *
	 * @param source [in] the source node of the edge
	 * @param target [in] the target node of the edge
	 * @param xadj [in] the xadj matrix of the CSR format
	 * @param adjncy [in] the adjncy matrix of the CSR format
	 * @return the index of matrix \a adjncy (or \a adjwgt) representing the
	 * 		   specified edge or -1 if not found
	 */
	inline static idx_t csr_edge_index(idx_t source, idx_t target,
														const idx_t *xadj, const idx_t *adjncy) {
		for (int neigh = xadj[source]; neigh < xadj[source + 1]; neigh++) {
			if (adjncy[neigh] == target) {
				return neigh;
			}
		}

		return -1;
	}

	/**
	 * Prints bytes in human readable format.
	 *
	 * @param bytes [in] the specified number of bytes
	 * @return the string value of \a bytes
	 */
	inline static std::string get_human_size(long bytes) {
		float kbytes =  bytes / 1000;
		float mbytes = kbytes / 1000;
		float gbytes = mbytes / 1000;

		std::stringstream strm;

		if (gbytes > 1) {
			strm << gbytes << " GB";

		} else if (mbytes > 1) {
			strm << mbytes << " MB";

		} else if (kbytes > 1) {
			strm << kbytes << " KB";

		} else {
			strm << bytes << " B";
		}

		return strm.str();
	}

	/**
	 * Computes the system's memory.
	 *
	 * @return the memory in bytes
	 */
	inline static long get_system_memory() {
		long mem;
#ifdef APPLE_OS
		int mib [] = { CTL_HW, HW_MEMSIZE };
		int64_t value = 0;
		size_t length = sizeof(value);

		sysctl(mib, 2, &value, &length, NULL, 0);

		mem = value;
#else
		long pages = sysconf(_SC_PHYS_PAGES);
		long page_size = sysconf(_SC_PAGE_SIZE);

		mem = pages * page_size;
#endif
		return mem;
	}

	inline static std::string get_time(long int start, long int end) {
		std::stringstream strm;
		strm << (end - start) / 1000 << " ms ("<< (end - start) / 1000000 <<" sec)";

        return strm.str();
	}

	virtual ~Util();
};

#endif /* UTIL_H_ */
