/**
 * This Source Code Form is subject to the terms of the Mozilla Public
 * License, v. 2.0. If a copy of the MPL was not distributed with this
 * file, You can obtain one at http://mozilla.org/MPL/2.0/.
 *
 * Copyright (C) 2013, 2014 Department of Informatics and
 * Telecommunications, National and Kapodistrian University of Athens
 *
 *  Created on: Dec 17, 2013
 *      Author: Charalampos Nikolaou <charnik@di.uoa.gr>
 */
#include <iostream>
#include <cstdlib>
#include <exception>
#include <string>

#include <boost/program_options.hpp>

#include "prcc8.h"

#include "PartGraphPicker.h"
#include "ConsistencyEngine.h"
#include "rcc8csp.h"
#include "MetisGraph.h"
#include "Util.h"

#include "ReasoningService.h"

int parse(int argc, const char **argv, cmd_options &);
void print_report(const std::vector<graph_info> &graphs);

inline
void
print_stats(const graph_info &g) {
	printf("# %10d\t%8.2f\t%12.2f\t%12.2f\t%-20s\t%s\n",
												g.consistent,
												util::get_time(g.partitioning_time + g.consistency_time),
												util::get_time(g.partitioning_time),
												util::get_time(g.consistency_time),
												g.part_desc.c_str(),
												g.filename);
}

extern bool valid_strategy(p_strat strategy);

int
main(int argc, const char **argv) {
	cmd_options opts;

	// parse command line
	int seen = parse(argc, argv, opts);

	// initialize environment
	ReasoningService::init_env();

	long int start_time = util::get_time();
	std::vector<graph_info> graphs;	/**< information per input graph */
	for (int i = 0; i < argc - seen; i++) {
		graph_info g(argv[seen + i]);
		ReasoningService srv(g, opts);

		if (opts.mem_check) { // check memory requirements and continue
			srv.estimate_memory_requirements();

			// add it to the array of input graphs
			graphs.push_back(g);
			continue;
		}

		if (opts.nparts == 1) { // run traditional consistency
			srv.check_traditional_consistency();

		} else {
			// create partition graph picker
			bool status = srv.partition();

			if (status && opts.print_pgraph) {
				srv.print_pgraph_to_file();
			}

			if (status && !opts.dryrun) {
				srv.check_consistency();
			}
		}

		// add it to the array of input graphs
		graphs.push_back(g);

		if (opts.quiet) print_stats(g);
	}
	long int end_time = util::get_time();

	if (!opts.quiet) print_report(graphs);

	if (graphs.size() > 1) {
		printf("Average processing time: %.2fs\n", util::get_time((end_time - start_time) / graphs.size()));
	}
	printf("Total processing time  : %.2fs\n", util::get_time(end_time - start_time));

	return EXIT_SUCCESS;
}

void print_report(const std::vector<graph_info> &graphs) {
	printf("  %-10s\t%-8s\t%-12s\t%-12s\t%-20s\t%s\n", "CONSISTENT", "TOTAL(s)", "PART_TIME(s)", "CONS_TIME(s)",
													   "STRATEGY", "FILENAME");

	for (unsigned int i = 0; i < graphs.size(); i++) {
		const graph_info &g = graphs[i];
		print_stats(g);
	}
}

int
parse(int argc, const char **argv, cmd_options &opts) {

	bool parse_error = false;
	std::string error_msg;
	int seen = 1;		/* arguments already seen (1 is for the program's name)*/
	std::string frmt;	/* network file format */

	boost::program_options::options_description desc("OPTIONS");
	desc.add_options()
			("parts,p", boost::program_options::value<int>(&opts.nparts), "number of partitions (if 1, then it runs traditional consistency; it suppresses options -d, -c, and -t)")
			("strat,s", boost::program_options::value<p_strat>(&opts.strategy), "partition strategy (one of \"balanced\" [default], \"blinkered\", \"min_fringe\", or \"min_inter\")")
			("memcheck,m", "check memory requirements for the requested arguments and exit")
			("pconly,z", "check only DPath-Consistency")
			("repartition,r", "after refining the network (using path-consistency), re-partition it and then check for consistency")
			("vol,o", "partition using communication volume as the objective function (default is cut-size)")
			("dry,d", "do not execute the consistency; only partition the input graph")
			("netviz_frmt,f", boost::program_options::value<std::string>(&frmt), "the file format of the network file to produce for vizualizing the computed partitioning graph (one of \"sif\", \"gml\", or \"gxmml\")")
			("sanity,c", "sanity check of the result based on the content of the partitioned network")
			("quiet,q", "prints only statistical information at the end of processing of each input network; otherwise a summary table will be printed at the end for all input networks")
			("threshold,t", "do not execute consistency when one of the partition's load is greater than 8000")
			("version,v", "print version information and exit")
			("license,l", "print license information and exit")
			("help,h", "print this message and exit");

	boost::program_options::variables_map args;
	try {
		boost::program_options::store(boost::program_options::parse_command_line(argc, argv, desc), args);
		boost::program_options::notify(args);

	} catch (std::exception &e) {
		error_msg = e.what();
		parse_error = true;

	}

	if (!parse_error && args.count("parts")) {
		seen += 2;
	}

	if (!parse_error && args.count("strat")) {
		if (!valid_strategy(opts.strategy)) {
			error_msg = "invalid selected strategy";
			parse_error = true;
		}

		seen += 2;
	}

	if (!parse_error && args.count("netviz_frmt")) {
		opts.pgraph_format = create_netviz_format(frmt);
		if (!valid_netviz_format(opts.pgraph_format)) {
			error_msg = "invalid selected network file format";
			parse_error = true;

		} else {
			opts.print_pgraph = true;

		}
		seen += 2;
	}


	if (args.count("dry")) {
		opts.dryrun = true;
		seen++;
	}

	if (args.count("vol")) {
		opts.obj_volume = true;
		seen++;
	}

	if (args.count("threshold")) {
		opts.threshold = true;
		seen++;
	}

	if (args.count("pconly")) {
		opts.pconly = true;
		seen++;
	}

	if (args.count("repartition")) {
		opts.repartition = true;
		seen++;
	}

	if (args.count("quiet")) {
		opts.quiet = true;
		seen++;
	}

	if (args.count("sanity")) {
		opts.sanity_check = true;
		seen++;
	}

	if (args.count("memcheck")) {
		opts.mem_check = true;
		seen++;
	}

	if (!parse_error && ((opts.dryrun && (argc - seen) == 0) ||
						 (opts.obj_volume && (argc - seen) == 0) ||
						 (opts.threshold && (argc - seen) == 0) ||
						 (opts.pconly && (argc - seen) == 0) ||
						 (opts.repartition && (argc - seen) == 0) ||
						 (opts.quiet && (argc - seen) == 0) ||
						 ((args.count("parts") || args.count("strat")) && (argc - seen) == 0)
																							)) {
		error_msg = "file argument is missing";
		parse_error = true;
	}

	if (args.count("help") || argc == 1 || parse_error) { // help
		if (parse_error) {
			std::cerr << PROG_NAME << ": " << error_msg << "\n\n";
		}
		std::cerr << "Usage: \n\t" << PROG_NAME << " [-vlh] [[-dmzrfctlq] [-p parts] [-s strategy] file...]\n\n";
		std::cerr << desc << "\n\n";
		std::cerr << "Author: Charalampos (Babis) Nikolaou <charnik@di.uoa.gr>\n";
		std::cerr << "Bugs: Please report any bugs to <charnik@di.uoa.gr>\n\n";
		std::cerr << LICENSE << "\n";
		exit(EXIT_FAILURE);
	}

	if (args.count("version")) { // version
		std::cerr << PRCC8_VERSION << "\n";
		exit(EXIT_SUCCESS);
	}

	if (args.count("license")) { // license
		std::cerr << PRCC8_VERSION << "\n";
		std::cerr << LICENSE << "\n";
		exit(EXIT_SUCCESS);
	}

	return seen;
}

