/**
 * This Source Code Form is subject to the terms of the Mozilla Public
 * License, v. 2.0. If a copy of the MPL was not distributed with this
 * file, You can obtain one at http://mozilla.org/MPL/2.0/.
 *
 * Copyright (C) 2013, 2014 Department of Informatics and
 * Telecommunications, National and Kapodistrian University of Athens
 *
 *  Created on: Dec 19, 2013
 *      Author: Charalampos Nikolaou <charnik@di.uoa.gr>
 */
#include "p_strat.h"

/* string aliases for partition strategies */
/* the array should be referenced by the enumeration type p_strat defined above */
const char *p_strat_alias[] = { "BA", "BL", "MF", "MI", "NONE"};

/* array of pointers to functions implementing the different partition strategies */
/* the array should be referenced by the enumeration type p_strat defined above */
void (*p_strat_impl[])(MetisGraph &, PartGraph *) = {pg_create_balanced,
													 pg_create_blinkered,
													 pg_create_minfringe,
													 pg_create_mininter};

void pg_pre_create(MetisGraph &mgraph, PartGraph *pgraph, long *start) {
	*start = util::get_time();
	for (int i = 0; i < mgraph.num_parts(); i++) {
		// create only non-empty partitions
		if (mgraph.part_size(i) > 0) {
			// create partition
			pgraph->create_partition(i, mgraph.graph(i));
		}
	}
}

void pg_post_create(MetisGraph &mgraph, PartGraph *pgraph, const edge_set &cut, long start) {
	pgraph->finalize(cut);

	// compute statistics
	pgraph->calc_stats();

	// set time
	pgraph->set_total_time(util::get_time() - start);
}

/**
 * Overload of operator>> for reading a p_strat enumeration type.
 *
 * @param in
 * @param unit
 * @return
 */
std::istream& operator>>(std::istream& in, p_strat& strategy) {
    std::string token;

    in >> token;

    if (token == BALANCE_NAME) {
    	strategy = BALANCED;

    } else if (token == BLINKERED_NAME) {
        strategy = BLINKERED;

    } else if (token == MIN_FRINGE_NAME) {
        strategy = MIN_FRINGE;

    } else if (token == MIN_INTERSECTION_NAME) {
    	strategy = MIN_INTERSECTION;

    } else {
    	strategy = PSTRAT_END;
    }

    return in;
}

p_strat& operator++(p_strat& strat) {
	assert(strat != PSTRAT_END);
	strat = static_cast<p_strat>(strat + 1);
	return strat;
}

p_strat operator++(p_strat& strat, int) {
	assert(strat != PSTRAT_END);
	++strat;
	return static_cast<p_strat>(strat - 1);
}

p_strat& operator--(p_strat& strat) {
	assert(strat != PSTRAT_BEGIN);
	strat = static_cast<p_strat>(strat - 1);
	return strat;
}

p_strat operator--(p_strat& strat, int) {
	assert(strat != PSTRAT_BEGIN);
	--strat;
	return static_cast<p_strat>(strat + 1);
}

/**
 * Creates a partitioning graph by trying to balance the final size of the partitions
 * participating in an edgecut. To do so, it checks whether the size of the source
 * partition after extended with the fringe of the target partition is less than
 * the size of the target partition after extended with the fringe of the left
 * partition. In such a case, we transfer the fringe of the right partition to the
 * left partition. Otherwise, we do the reverse transfer.
 *
 * Prerequisite: the initial graph should have already been partitioned by MeTiS.
 *
 * @param [in] mgraph the instance of the MetisGraph that partitioned the initial graph
 * @param [out] pgraph the created partition (should have been allocated prior to the call)
 */
void pg_create_balanced(MetisGraph &mgraph, PartGraph *pgraph) {
	long start;
	pg_pre_create(mgraph, pgraph, &start);

	edge_set edgecut;
	edgecut_map edges_per_edgecut;
	edgecut = mgraph.edgecut(edgecut, edges_per_edgecut);
	//cout << "Edges on edgecut: " << edges_per_edgecut.size() << endl;
	set<pair<idx_t, idx_t> >::const_iterator it;
	for (it = edgecut.begin(); it != edgecut.end(); ++it) {
		set<idx_t> fringe_spart, fringe_tpart;
		edge_list ind_graph;

		part_node& snode = pgraph->get_partition(it->first);
		part_node& tnode = pgraph->get_partition(it->second);

		// get the subset of the fringe of the target partition that is connected
		// to the source partition
		fringe_tpart = mgraph.fringe_for_edgecut(it->second, it->first, fringe_tpart);

		// get the subset of the fringe of the source partition that is connected
		// to the target partition
		fringe_spart = mgraph.fringe_for_edgecut(it->first, it->second, fringe_spart);

		part_node *to_keep = &tnode;					/* the partition to keep intact */
		part_node *to_modify = &snode;				/* the partition to modify */
		set<idx_t> *fringe_tocpy = &fringe_tpart;	/* the fringe to copy */

		if (tnode.size() + fringe_spart.size() == snode.size() + fringe_tpart.size()) { // the sizes are equal
			// break the tie looking at the partition sizes
			if (tnode.size() < snode.size()) { // move to the target
				to_keep = &snode;
				to_modify = &tnode;
				fringe_tocpy = &fringe_spart;

			} else if (tnode.size() == snode.size()) {
//				std::cout << "pg_create_balanced: cannot break tie even by looking at the partition sizes\n";
			}

		} else if (tnode.size() + fringe_spart.size() < snode.size() + fringe_tpart.size()) {
			to_keep = &snode;
			to_modify = &tnode;
			fringe_tocpy = &fringe_spart;
		}

		// get the induced graph of this set of vertices and augment the source partition
		ind_graph = mgraph.induced_graph(to_keep->id(), *fringe_tocpy, ind_graph);

		// keep track which node gives its induced graph
		to_modify->add_to_senders(*to_keep);

		// get the edges for the edgecut between the source and target partition and
		// augment the *appropriate* partition
		to_modify->augment_graph(edges_per_edgecut.at(*it));

		// update the MeTiS graph
		mgraph.augment_graph(to_modify->id(), edges_per_edgecut.at(*it));

		if (ind_graph.size() > 0) {
			to_modify->augment_graph(ind_graph, true);

			// update the MeTiS graph
			mgraph.augment_graph(to_modify->id(), ind_graph);

			// only in this case we should mark the dependency of the two partitions
			pgraph->add_or_update_edge(to_modify->id(), to_keep->id(), ind_graph);
		}

		// update the fringe maintained by MeTiS
		edge_set updated_cuts;
		mgraph.augment_fringe_for_edgecut(to_modify->id(), to_keep->id(), edgecut, *fringe_tocpy, updated_cuts);

	}

	pg_post_create(mgraph, pgraph, edgecut, start);
}

/**
 * Creates a partitioning graph in a fixed/blinkered way by copying nodes, edges, and
 * edgecuts between two partitions by looking only at their ids. The graph elements
 * are copied and transferred (case of edgecuts) always to the partition with the
 * lowest id.
 *
 * Prerequisite: the initial graph should have already been partitioned by MeTiS.
 *
 * @param [in] mgraph the instance of the MetisGraph that partitioned the initial graph
 * @param [out] pgraph the created partition (should have been allocated prior to the call)
 */
void pg_create_blinkered(MetisGraph &mgraph, PartGraph *pgraph) {
	long start;
	pg_pre_create(mgraph, pgraph, &start);

	edge_set edgecut;
	edgecut_map edges_per_edgecut;
	edgecut = mgraph.edgecut(edgecut, edges_per_edgecut);
	//cout << "Edges on edgecut: " << edges_per_edgecut.size() << endl;
	set<pair<idx_t, idx_t> >::const_iterator it;
	for (it = edgecut.begin(); it != edgecut.end(); ++it) {
		set<idx_t> fringe_part;
		edge_list ind_graph;
//std::cout << "1" << "\n";
		part_node& snode = pgraph->get_partition(it->first);
		part_node& tnode = pgraph->get_partition(it->second);

		// get the subset of the fringe of the target partition that is connected
		// to the source partition
		fringe_part = mgraph.fringe_for_edgecut(it->second, it->first, fringe_part);
//std::cout << "2" << "\n";
		// get the induced graph of this set of vertices and augment the source partition
		ind_graph = mgraph.induced_graph(it->second, fringe_part, ind_graph);
//std::cout << "3" << "\n";
		// keep track which node gives its induced graph
		snode.add_to_senders(tnode);
//std::cout << "4" << "\n";
		// get the edges for the edgecut between the source and target partition and
		// augment the source partition
		snode.augment_graph(edges_per_edgecut.at(*it));
//std::cout << "5" << "\n";
		// update the MeTiS graph
		mgraph.augment_graph(snode.id(), edges_per_edgecut.at(*it));
//std::cout << "6" << "\n";
		if (ind_graph.size() > 0) {
			snode.augment_graph(ind_graph, true);
//std::cout << "7" << "\n";
			// update the MeTiS graph
			mgraph.augment_graph(snode.id(), ind_graph);
//std::cout << "8" << "\n";
			// only in this case we should mark the dependency of the two partitions
			pgraph->add_or_update_edge(it->first, it->second, ind_graph);
		}
//std::cout << "9" << "\n";
		// update the fringe maintained by MeTiS
		edge_set updated_cuts;
		mgraph.augment_fringe_for_edgecut(snode.id(), it->second, edgecut, fringe_part, updated_cuts);
//std::cout << "10" << "\n";
	}

	pg_post_create(mgraph, pgraph, edgecut, start);
}

/**
 * Creates a partitioning graph by looking at the sizes of the fringes that are
 * to be copied between two partitions connected through an edgecut. This strategy
 * tries to minimize the fringe size to be copied, with the aim of minimizing the
 * memory requirements for a specific pair of partitions. Ties are broken by looking
 * at the sizes of the two partitions, always copying the nodes of the bigger partition
 * to the smaller partition.
 *
 * Prerequisite: the initial graph should have already been partitioned by MeTiS.
 *
 * @param [in] mgraph the instance of the MetisGraph that partitioned the initial graph
 * @param [out] pgraph the created partition (should have been allocated prior to the call)
 */
void pg_create_minfringe(MetisGraph &mgraph, PartGraph *pgraph) {
	long start;
	pg_pre_create(mgraph, pgraph, &start);

	edge_set edgecut;
	edgecut_map edges_per_edgecut;
	edgecut = mgraph.edgecut(edgecut, edges_per_edgecut);
	//cout << "Edges on edgecut: " << edges_per_edgecut.size() << endl;
	set<pair<idx_t, idx_t> >::const_iterator it;
	for (it = edgecut.begin(); it != edgecut.end(); ++it) {
		set<idx_t> fringe_spart, fringe_tpart;
		edge_list ind_graph;

		part_node& snode = pgraph->get_partition(it->first);
		part_node& tnode = pgraph->get_partition(it->second);

		// get the subset of the fringe of the target partition that is connected
		// to the source partition
		fringe_tpart = mgraph.fringe_for_edgecut(it->second, it->first, fringe_tpart);

		// get the subset of the fringe of the source partition that is connected
		// to the target partition
		fringe_spart = mgraph.fringe_for_edgecut(it->first, it->second, fringe_spart);

		part_node *to_keep = &tnode;				/* the partition to keep intact */
		part_node *to_modify = &snode;				/* the partition to modify */
		set<idx_t> *fringe_tocpy = &fringe_tpart;	/* the fringe to copy */

		if (fringe_spart.size() == fringe_tpart.size()) { // the sizes are equal
			// break the tie looking at the partition sizes
			if (tnode.size() < snode.size()) { // move to the target
				to_keep = &snode;
				to_modify = &tnode;
				fringe_tocpy = &fringe_spart;

			} else if (tnode.size() == snode.size()) {
//				std::cout << "pg_create_minfringe: cannot break tie even by looking at the partition sizes\n";
			}

		} else if (fringe_spart.size() < fringe_tpart.size()) {
			to_keep = &snode;
			to_modify = &tnode;
			fringe_tocpy = &fringe_spart;
		}

		// keep track which node gives its induced graph
		to_modify->add_to_senders(*to_keep);

		// get the edges for the edgecut between the source and target partition and
		// augment the *appropriate* partition
		to_modify->augment_graph(edges_per_edgecut.at(*it));

		// update the MeTiS graph
		mgraph.augment_graph(to_modify->id(), edges_per_edgecut.at(*it));

		// get the induced graph of this set of vertices and augment the source partition
		ind_graph = mgraph.induced_graph(to_keep->id(), *fringe_tocpy, ind_graph);
		if (ind_graph.size() > 0) {
			to_modify->augment_graph(ind_graph, true);

			// update the MeTiS graph
			mgraph.augment_graph(to_modify->id(), ind_graph);

			// only in this case we should mark the dependency of the two partitions
			pgraph->add_or_update_edge(to_modify->id(), to_keep->id(), ind_graph);
		}

		// update the fringe maintained by MeTiS
		edge_set updated_cuts;
		mgraph.augment_fringe_for_edgecut(to_modify->id(), to_keep->id(), edgecut, *fringe_tocpy, updated_cuts);

	}

	pg_post_create(mgraph, pgraph, edgecut, start);
}

/**
 * Creates a partitioning graph by looking at the induced graph corresponding to
 * the fringes of two partitions connected through an edgecut. This strategy tries
 * to minimize the size of edges that are copied between two partitions, and hence
 * it tries to minimize their in between intersection. This strategy is particularly
 * useful in cases of partitions that are completely disconnected, i.e., they are
 * composed of only (or very many) disconnected nodes. Ties are broken by looking at
 * the sizes of the fringes of the two partitions, always copying the fringe of the
 * bigger partition to the fringe of the smaller partition. Ties are broken according
 * to \a pg_create_minnodes(), i.e., looking at the partition sizes.
 *
 * Prerequisite: the initial graph should have already been partitioned by MeTiS.
 *
 * @param [in] mgraph the instance of the MetisGraph that partitioned the initial graph
 * @param [out] pgraph the created partition (should have been allocated prior to the call)
 */
void pg_create_mininter(MetisGraph &mgraph, PartGraph *pgraph) {
	long start;
	pg_pre_create(mgraph, pgraph, &start);

	edge_set edgecut;
	edgecut_map edges_per_edgecut;
	edgecut = mgraph.edgecut(edgecut, edges_per_edgecut);
	//cout << "Edges on edgecut: " << edges_per_edgecut.size() << endl;
	set<pair<idx_t, idx_t> >::const_iterator it;
	for (it = edgecut.begin(); it != edgecut.end(); ++it) {
		set<idx_t> fringe_spart, fringe_tpart;
		edge_list ind_graph_spart, ind_graph_tpart;

		part_node& snode = pgraph->get_partition(it->first);
		part_node& tnode = pgraph->get_partition(it->second);

		// get the subset of the fringe of the target partition that is connected
		// to the source partition
		fringe_tpart = mgraph.fringe_for_edgecut(it->second, it->first, fringe_tpart);
		// get the corresponding induced graph
		ind_graph_tpart = mgraph.induced_graph(it->second, fringe_tpart, ind_graph_tpart);

		// get the subset of the fringe of the source partition that is connected
		// to the target partition
		fringe_spart = mgraph.fringe_for_edgecut(it->first, it->second, fringe_spart);
		// get the corresponding induced graph
		ind_graph_spart = mgraph.induced_graph(it->first, fringe_spart, ind_graph_spart);

		part_node *to_keep = &tnode;						/* the partition to keep intact */
		part_node *to_modify = &snode;						/* the partition to modify */
		edge_list *ind_graph_tocpy = &ind_graph_tpart;		/* the induced graph to copy */
		set<idx_t> *fringe_tocpy = &fringe_tpart;			/* the fringe a subset of which might be copied */

		if (ind_graph_spart.size() == ind_graph_tpart.size()) { // the sizes are equal
//			std::cout << "pg_create_mininter: breaking tie by looking at the fringe sizes\n";
			// break the tie looking at the fringe sizes
			if (fringe_spart.size() < fringe_tpart.size()) { // move to the target
				to_keep = &snode;
				to_modify = &tnode;
				ind_graph_tocpy = &ind_graph_spart;
				fringe_tocpy = &fringe_spart;

			} else if (fringe_spart.size() == fringe_tpart.size()) { // fringes are equal
				// break the tie looking at the partition sizes
				if (tnode.size() < snode.size()) { // move to the target
					to_keep = &snode;
					to_modify = &tnode;
					ind_graph_tocpy = &ind_graph_spart;
					fringe_tocpy = &fringe_spart;

				} else if (tnode.size() == snode.size()) {
//					std::cout << "pg_create_mininter: cannot break tie even by looking at the partition sizes\n";
				}
			}

		} else if (ind_graph_spart.size() < ind_graph_spart.size()) {
			to_keep = &snode;
			to_modify = &tnode;
			ind_graph_tocpy = &ind_graph_spart;
			fringe_tocpy = &fringe_spart;

		}

		// keep track which node gives its induced graph
		to_modify->add_to_senders(*to_keep);

		// get the edges for the edgecut between the source and target partition and
		// augment the *appropriate* partition
		to_modify->augment_graph(edges_per_edgecut.at(*it));

		// update the MeTiS graph
		mgraph.augment_graph(to_modify->id(), edges_per_edgecut.at(*it));

		if (ind_graph_tocpy->size() > 0) {
//			std::cout << "processing edgecut " << it->first << " -- " << it->second << "\n";
			to_modify->augment_graph(*ind_graph_tocpy, true);

			// update the MeTiS graph
			mgraph.augment_graph(to_modify->id(), *ind_graph_tocpy);

			// only in this case we should mark the dependency of the two partitions
			pgraph->add_or_update_edge(it->first, it->second, *ind_graph_tocpy);

		}
		else {
//			std::cout << "processing edgecut " << it->first << " -- " << it->second << " (not adding)\n";
		}

		// update the fringe maintained by MeTiS
		edge_set updated_cuts;
		mgraph.augment_fringe_for_edgecut(to_modify->id(), to_keep->id(), edgecut, *fringe_tocpy, updated_cuts);

	}

	pg_post_create(mgraph, pgraph, edgecut, start);
}

