/**
 * This Source Code Form is subject to the terms of the Mozilla Public
 * License, v. 2.0. If a copy of the MPL was not distributed with this
 * file, You can obtain one at http://mozilla.org/MPL/2.0/.
 *
 * Copyright (C) 2013, 2014 Department of Informatics and
 * Telecommunications, National and Kapodistrian University of Athens
 *
 *  Created on: Dec 17, 2013
 *      Author: Charalampos Nikolaou <charnik@di.uoa.gr>
 */
#ifndef PARTNODE_H_
#define PARTNODE_H_

#include "datastruct.h"
#include "metis/metis.h"

#include <vector>

class part_node;
typedef boost::unordered_map<idx_t, part_node> partition_map;

class part_node {
private:

	int pid;				/**< the id of the partition produced by Metis */
	bool isvirtual;			/**< true when this partition have not been constructed originally by MeTiS */

	std::set<idx_t> nodes;		/**< set of graph nodes */
	edge_set edges;				/**< set of graph edges */
	std::set<idx_t> take_from;  /**< id of partitions that were chosen to give their induced graphs to this partition */
	int num_shared_edges;		/**< number of shared edges (always appear first in the \a edges list */
	int init_part_nodesz;		/**< the initial number of nodes of the partition by MeTiS */

//	part_node(part_node &n) {
//
//	}

public:
	part_node(int id, bool virt, const edge_set &graph) {
		pid = id;
		isvirtual = virt;
		nodes = std::set<idx_t>();
		edges = edge_set();
		take_from = std::set<idx_t>();

		init_part_nodesz = 0;
		num_shared_edges = 0;

		edge_set::const_iterator it;
		for (it = graph.begin(); it != graph.end(); ++it) {
			edges.insert(*it);

			nodes.insert(it->first);
			nodes.insert(it->second);
		}

		init_part_nodesz = nodes.size();
	}

	inline void add_to_senders(const part_node &sender) {
		take_from.insert(sender.id());
	}

	inline bool receives_from(const part_node & other) const {
		return take_from.find(other.id()) != take_from.end();
	}

	inline void set_id(int id) {
		pid = id;
	}

	inline void set_virtual(bool virt) {
		isvirtual = virt;
	}

	inline int id() const {
		return pid;
	}

	inline bool is_disconnected() const {
		return edges.empty();
	}

	inline int size() const {
		return nodes.size();
	}

	inline bool empty() const {
		return nodes.size() == 0;
	}

	inline int init_size() const {
		return init_part_nodesz;
	}

	inline int get_num_shared_edges() const {
		return num_shared_edges;
	}

	inline bool is_virtual() const {
		return isvirtual;
	}

	/**
	 * @return the set of nodes
	 */
	inline const std::set<idx_t>& get_nodes() const {
		 return nodes;
	}

	/**
	 * @return the list of edges
	 */
	inline const edge_set& get_edges() const {
		 return edges;
	}

	inline edge_set& get_edges() {
		 return edges;
	}

	/**
	 * Augment the graph with the given set of edges.
	 *
	 * @param new_edges the set of edges to add
	 * @param shared true when the edges specified in \a new_edges appear in other partitions
	 * 		  as well
	 */
	void augment_graph(const edge_list &new_edges, bool shared = false) {
		edge_list::const_iterator it;
		for (it = new_edges.begin(); it != new_edges.end(); ++it) {
			if (edges.insert(*it).second && shared) {
				num_shared_edges++; // update their number
			}

			nodes.insert(it->first);
			nodes.insert(it->second);
		}
	}

	inline void print() const {
		std::cout << "[ ";
		std::set<idx_t>::const_iterator itn;
		for (itn = nodes.begin(); itn != nodes.end(); ++itn) {
			std::cout << *itn << " ";
		}
		std::cout << "]\n";

		edge_set::const_iterator it;
		for (it = edges.begin(); it != edges.end(); ++it) {
			std::cout << "(" << it->first << ", " << it->second << ")\n";
		}
	}
};


#endif /* PARTNODE_H_ */
