/**
 * This Source Code Form is subject to the terms of the Mozilla Public
 * License, v. 2.0. If a copy of the MPL was not distributed with this
 * file, You can obtain one at http://mozilla.org/MPL/2.0/.
 *
 * Copyright (C) 2013, 2014 Department of Informatics and
 * Telecommunications, National and Kapodistrian University of Athens
 *
 *  Created on: Jan 18, 2014
 *      Author: Charalampos Nikolaou <charnik@di.uoa.gr>
 */
#include <iostream>
#include <cstdio>
#include "rcc8.h"
#include "prcc8.h"

#include <boost/graph/adjacency_list.hpp>
#include <boost/graph/plod_generator.hpp>
#include <boost/graph/connected_components.hpp>
#include <boost/graph/random.hpp>

#include <boost/random/linear_congruential.hpp>

#include <boost/random/mersenne_twister.hpp>
#include <boost/random/uniform_smallint.hpp>
#include <boost/random/variate_generator.hpp>

#include <boost/program_options.hpp>

typedef boost::adjacency_list<boost::listS, boost::vecS, boost::undirectedS> Graph;
typedef boost::plod_iterator<boost::minstd_rand, Graph> SFGen;
typedef boost::uniform_smallint<> small_distribution;
typedef boost::variate_generator<boost::mt19937&, small_distribution> vgenerator;

#define NPRELS_SIZE	76

/* the set of the 79 hard relations of the subset of RCC-8 relations NP-8 */
RELTYPE nprels[NPRELS_SIZE] = {40, 41, 42, 43, 48, 49, 50, 51, 56, 57, 58, 59, 72, 73, 74, 75, 80, 81, 82, 83, 88, 89, 90, 91, 104, 105, 106, 107, 112, 113, 114, 115, 120, 121, 122, 123, 146, 147, 168, 169, 170, 171, 176, 177, 178, 179, 184, 185, 186, 187, 194, 195, 200, 201, 202, 203, 208, 209, 210, 211, 216, 217, 218, 219, 232, 233, 234, 235, 240, 241, 242, 243, 248, 249, 250, 251};

/**
 * Parse command line arguments
 */
void parse(int argc, const char **argv);

/**
 * Finds all connected components of an undirected graph
 * and connects them randomly.
 *
 * @param g an undirected graph
 */
void connect_components(Graph &g);

/**
 * Creates and return a generator of values in the specified range.
 *
 * @param min the minimum value that will be generator by this generator
 * @param max the maximum value that will be generator by this generator
 * @return the generator object
 */
inline
vgenerator
get_generator(int min, int max) {
	small_distribution dist(min, max);
	boost::mt19937 engine(time(NULL));
	vgenerator gen(engine, dist);

	return gen;
}

/**
 * Given an undirected graph, it calculates the average degree.
 *
 * @param g
 * @return
 */
inline
float
avg_degree(Graph &g) {
	float degs = 0;
	for (unsigned int i = 0; i < boost::num_vertices(g); i++) {
		degs += boost::degree(i, g);
	}

	return degs / boost::num_vertices(g);

}

/* filename template when generating multiple instances of power-law networks */
#define FILE_TEMPL "plaw-n%d-i%d.csp"

/* parameters for generating power-law networks */
int nnodes;				/**< the number of nodes in the generated network */
double alpha = 2.5;		/**< the parameter alpha: beta/x^alpha */
double beta = 1.0;		/**< the parameter beta: beta/x^alpha */
int instncs = 1;

int
main(int argc, const char **argv) {
	// parse command line
	parse(argc, argv);

	srand (time(NULL));

	for (int i = 0; i < instncs; i++) {
		boost::minstd_rand gen(time(NULL));					/**< generator used for connected nodes */
		small_distribution rel_dist(0, NPRELS_SIZE - 1);	/**< distribution for generating RCC-8 relations from the set NP-8 */
		boost::mt19937 engine(time(NULL));
		vgenerator relgen(engine, rel_dist);				/**< generator of RCC-8 relations */

		FILE *out;											/**< the output file descriptor to write the network */
		if (instncs > 1) {
			char filename[BUFSIZ];
			sprintf(filename, FILE_TEMPL, nnodes, i);
			if ((out = fopen(filename, "w")) == NULL) {
				perror("cannot open file for writing");
				exit(-1);
			}
		} else {
			out = stdout;
		}

		// create a random graph with the specified number of nodes
		Graph g(SFGen(gen, nnodes, alpha, beta), SFGen(), nnodes);

		// ensure connectedness of components
		connect_components(g);

		// print header
		fprintf(out, "%d #N%d-A%.2f-B%.2f\n", nnodes - 1, nnodes, alpha, beta);

		// print labeled edges
		for (int i = 0; i < nnodes; i++) {
			for (int j = i + 1; j < nnodes; j++) {
				assert (boost::edge(i, j, g).second == boost::edge(j, i, g).second);
				if (boost::edge(i, j, g).second) { // if this edge is present
					// print labeled edge
					fprintf(out, " %d %d %s\n", i, j, print_relation(nprels[relgen()]).c_str());
				}
			}
		}

		// print footer
		fprintf(out, ".\n");

		if (instncs > 1) {
			fclose(out);
			out = NULL;
		}

		//std::cerr << "avg degree: " << avg_degree(g) << "\n";

	}

	return EXIT_SUCCESS;
}

void
connect_components(Graph &g) {
	std::vector<int> component(boost::num_vertices(g));
	int ncomp = boost::connected_components(g, &component[0]);
	//std::cout << "found " << ncomp << " connected components\n";

	while (ncomp > 1) {
		// create generator
		small_distribution dist(0, ncomp - 1);
		boost::mt19937 engine(time(NULL));
		boost::variate_generator<boost::mt19937&, small_distribution> gen(engine, dist);

		int c1 = -1, c2 = -1;
		do { // choose randomly the components to connect
			c1 = gen();
			c2 = gen();
		} while (c1 == c2);

		// compute the sets of vertices belonging to these components
		std::vector<int> c1_nodes, c2_nodes;
		for (unsigned int v = 0; v < component.size(); v++) {
			if (component[v] == c1) {
				c1_nodes.push_back(v);

			} else {

				if (component[v] == c2) {
					c2_nodes.push_back(v);
				}
			}
		}

		vgenerator gen_c1 = get_generator(0, c1_nodes.size() - 1);
		vgenerator gen_c2 = get_generator(0, c2_nodes.size() - 1);

		// pick random nodes
		int u = c1_nodes[gen_c1()];
		int v = c2_nodes[gen_c2()];

		boost::add_edge(u, v, g);
		assert (boost::edge(u, v, g) == boost::edge(v, u, g));

		// re-compute connected components
		ncomp = boost::connected_components(g, &component[0]);
	}
}

void
parse(int argc, const char **argv) {
	boost::program_options::options_description desc("OPTIONS");
	desc.add_options()
			("instances,i", boost::program_options::value<int>(&instncs),
						"number of instances to generate (these will go to files with names plaw-n<nodes>-i<inst>.csp) [default is 1 and is written to stdout")
			("alpha,a", boost::program_options::value<double>(&alpha),
						"alpha parameter: a floating point number usually in the range [1.9, 2.7] (default is 2.5)")
			("beta,b", boost::program_options::value<double>(&beta),
					   "beta parameter: a floating point number in the range [1.0, -) (default is 1.0)")
			("version,v", "print version information and exit")
			("license,l", "print license information and exit")
			("help,h", "print this message and exit");

	bool parse_error = false;
	std::string error_msg;
	int seen = 1;				/* arguments already seen (1 is for the program's name)*/

	boost::program_options::variables_map args;
	try {
		boost::program_options::store(boost::program_options::parse_command_line(argc, argv, desc), args);
		boost::program_options::notify(args);

	} catch (std::exception &e) {
		error_msg = e.what();
		parse_error = true;

	}

	if (!parse_error && args.count("instances")) {
		seen += 2;

		// check validity
		if (instncs < 1.0) {
			error_msg = "invalid number of instances";
			parse_error = true;
		}
	}

	if (!parse_error && args.count("alpha")) {
		seen += 2;

		// check validity
		if (alpha < 1.0) {
			error_msg = "invalid alpha parameter";
			parse_error = true;
		}
	}

	if (!parse_error && args.count("beta")) {
		seen += 2;

		// check validity
		if (beta < 1.0) {
			error_msg = "invalid beta parameter";
			parse_error = true;
		}
	}

	if (!parse_error && (argc == 1 ||
						 ((args.count("alpha") || args.count("beta")) && (argc - seen) == 0)
																							)) {
		error_msg = "unspecified number of nodes";
		parse_error = true;

	} else {
		// get number of nodes
		nnodes = std::atoi(argv[seen]);
	}

	if (args.count("help") || parse_error) { // help
		if (parse_error) {
			std::cerr << GEN_NAME << ": " << error_msg << "\n\n";
		}
		std::cerr << "Usage: \n\t" << GEN_NAME << " [-vlh] [[-a alpha] [-b beta] nodes]\n\n";
		std::cerr << desc << "\n\n";
		std::cerr << "Author: Charalampos (Babis) Nikolaou <charnik@di.uoa.gr>\n";
		std::cerr << "Bugs: Please report any bugs to <charnik@di.uoa.gr>\n\n";
		std::cerr << LICENSE << "\n";
		exit(EXIT_FAILURE);
	}

	if (args.count("version")) { // version
		std::cerr << PGEN_VERSION << "\n";
		exit(EXIT_SUCCESS);
	}

	if (args.count("license")) { // license
		std::cerr << PGEN_VERSION << "\n";
		std::cerr << LICENSE << "\n";
		exit(EXIT_SUCCESS);
	}
}

