/**
 * This Source Code Form is subject to the terms of the Mozilla Public
 * License, v. 2.0. If a copy of the MPL was not distributed with this
 * file, You can obtain one at http://mozilla.org/MPL/2.0/.
 *
 * Copyright (C) 2013, 2014 Department of Informatics and
 * Telecommunications, National and Kapodistrian University of Athens
 *
 *  Created on: Dec 28, 2013
 *      Author: Charalampos Nikolaou <charnik@di.uoa.gr>
 */
#ifndef RCC8_H_
#define RCC8_H_

#include "metis/metis.h"
#include "datastruct.h"

#include <cmath>

#include "boost/tuple/tuple.hpp"

/* increment a bitset by one */
#define BITSET_INC(b)   for (boost::dynamic_bitset<>::size_type i = 0, len = (b).size(); i < len && (((b)[i] ^= 1) == 0); i++)\
	;

#define FREE_SPLIT(s)	if ((s) != NULL) { delete (s)->it; delete (s)->msb_it; (s)->edges.clear(); (s) = NULL; }
#define FREE_BSPLIT(s)	if ((s) != NULL) { delete (s)->it; (s)->edges.clear(); (s)->rels.clear(); (s) = NULL; }

/* copied from rcc8.h */
#define RELTYPE short int

#define RELTYPESIZE sizeof(RELTYPE)

/* Coding of the RCC8 relations */
/* coded as binary, assuming int 8bit or more */
#define DC                1
#define EC                2
#define PO                4
#define TPP               8
#define NTPP             16
#define TPPI             32
#define NTPPI            64
#define EQ              128
#define DALL         	255

/* Number of possible disjunctions (including 0 for empty Relation) */
#define MAXSET          256

/* Number of basic relations */
#define MAXREL          8

/* an RCC-8 constraint is an edge (idx_t, idx_t) with an RCC-8 relation (RELTYPE) */
typedef boost::tuple<idx_t, idx_t, RELTYPE> rcc8_con;

extern RELTYPE horn_split[MAXSET][4];
extern RELTYPE base_split[MAXSET][9];
extern short int vb_weight[MAXREL];

/**
 * A split_set structure is a very compact representation of all possible combinations of
 * how a set of RCC-8 edges can be split in disjunctions from the set of horn relations.
 * Compactness is gained from the fact that any RCC-8 relation can be split to a
 * disjunction of horn relations with two or three disjuncts. For example, the RCC-8
 * relation {DC, EC, NTPP, NTPPI, EQ} can be split to the following disjunction composed
 * of three horn RCC-8 relations: {DC EC NTPP} \/ {NTPPI} \/ {EQ}.
 *
 *          it : this is an iterator over a bitset of size equal to the number of edges
 *               participating in this split; this iterator is always incremented by one
 *      msb_it : this is an iterator over a bitset of size equal to the number of edges
 *               that should be split in horn relations with three disjuncts; if any bit
 *               of this iterator is set, then the corresponding edge will get the third
 *               horn relation
 *       edges : this is the list of edges that are split; those edges that are split to
 *               three horn relations, are always stored in the beginning of this list
 */
typedef struct split_set {
	bit_set* it;
	bit_set* msb_it;
	edge_list edges;
	bool at_end;

	split_set() {
		it = NULL;
		msb_it = NULL;

		at_end = false;
		edges = edge_list();
	}

	/**
	 * Check whether the iterator is at the end. The clients should
	 * call this function to determine that there is no other split
	 * to check and not the \a has_next() function. The usage should
	 * be as follows for a split_set pointer \a s:
	 *
	 * while (!s->end()) {
	 * 		// do something with s
	 * 		// ...
	 *
	 * 		// increment the split iterator at the end of the loop
	 * 		s->incr();
	 * }
	 *
	 * @return
	 */
	inline bool end() const {
		return at_end;
	}

	/**
	 * Checks whether there is another split to iterate over.
	 *
	 * @return
	 */
	inline bool _has_next() const {
		if (it->count() == it->size()) { // all bits set
			boost::dynamic_bitset<>::size_type sz = msb_it->size();
			if (sz == 0 || msb_it->count() == sz) { // no three-split edges or all bits set
				return false;
			}
		}

		return true;
	}

	/**
	 * Increment the iterator by one, i.e., representing the next
	 * split of the set of edges. If the function is called when
	 * the iterator is at its last position, then there is no
	 * increment under the hood, but we mark that we are at the
	 * end.
	 */
	inline void incr() {
		boost::dynamic_bitset<>::size_type sz = it->size();

		if (!_has_next()) {
			at_end = true;
			return ;
		}

		if (it->count() < sz) {
			BITSET_INC(*it);

		} else { // all bits of "it" are set; has_next() has returned true and
				 // hence, we should increment the iterator over three-split
				 // edges and reset the main counter

			// increase the counter for the other iterator
			BITSET_INC(*msb_it);

			delete it;

			// decrease the number of bits by the number of bits that are set in msb_it;
			// the corresponding edges for the set bits should take the third split and
			// they should not be involved in the counting of the two-split edges
			it = new bit_set(edges.size() - msb_it->count());
		}
	}

	/**
	 * Returns the position of the split in the horn_split array (either 1, 2, or 3)
	 * for a specific edge identified by its index position, according to the current
	 * state of the split.
	 *
	 * @param idx the index position of the edge in the edges list
	 * @return the position of the split in the horn_split array (possible values are: 1, 2, and 3)
	 */
	inline short int get_split_pos(boost::dynamic_bitset<>::size_type idx) {
		if (msb_it->size() == 0  || // if there is no three-split relation OR
			msb_it->count() == 0 || // we haven't reached the counter yet  OR (this is subsumed by the next condition, but we leave it for readability)
			idx < it->size()) { 	// the requested index is referred to two-split edges
			return (*it)[idx] + 1;

		} // some bit of msb_it is set

		// get its index with respect to msb_it
		boost::dynamic_bitset<>::size_type msb_idx = msb_it->size() - (edges.size() - idx);
		if ((*msb_it)[msb_idx] == 1) { // it is set
			return 3;

		} // it is not set, look it up in "it"

		// count the number of zeros up to msb_idx
		boost::dynamic_bitset<>::size_type zeros = 0;
		for (boost::dynamic_bitset<>::size_type i = msb_it->size() - 1; i >= msb_idx; --i) {
			if ((*msb_it)[i] == 0) {
				zeros++;
			}
		}

		return (*it)[it->size() - zeros] + 1;
	}

	inline long double get_max_value() const {
		return pow(2, edges.size() - msb_it->size()) * pow(3, msb_it->size());
	}

} split_set;





typedef struct bsplit_set {
	bit_set* it;
	edge_list edges;
	std::list<RELTYPE> rels;
	bool at_end;

	bsplit_set() {
		it = NULL;
		edges = edge_list();
		rels = std::list<RELTYPE>();
		at_end = false;
	}

	inline bool end() {
		return at_end;
	}

	/**
	 * Checks whether there is another split to check.
	 *
	 * @return
	 */
	inline bool _has_next() const {
		boost::dynamic_bitset<>::size_type idx = 0;
		std::list<RELTYPE>::const_reverse_iterator itl;
		for (itl = rels.rbegin(); itl != rels.rend(); ++itl) {
			bit_set mask = bit_set(it->size());
			mask.set(0); mask.set(1); mask.set(2);

			unsigned long idx_counter = (((mask <<= idx) & (*it)) >>= idx).to_ulong();

			if (idx_counter + 1 == (unsigned long) base_split[*itl][0]) {
				idx += 3;

			} else {
				return true;
			}
		}

		return false;
	}

	/**
	 * Increment the iterator by one, i.e., representing the next
	 * split of the set of edges.
	 *
	 * Prerequisite: the function \a has_next() should have been called
	 * and returned true.
	 */
	inline void incr() {
		if (!_has_next()) {
			at_end = true;
			return ;
		}

		boost::dynamic_bitset<>::size_type idx = 0;
		std::list<RELTYPE>::const_reverse_iterator itl;
		for (itl = rels.rbegin(); itl != rels.rend(); ++itl) {
			bit_set mask = bit_set(it->size());
			mask.set(0); mask.set(1); mask.set(2);

			unsigned long idx_counter = (((mask <<= idx) & (*it)) >>= idx).to_ulong();

			if (idx_counter + 1 == (unsigned long) base_split[*itl][0]) {
				idx += 3;

			} else {
				idx_counter++;
				bit_set newval = bit_set(3, idx_counter);

				(*it)[idx] = newval[0];
				(*it)[idx+1] = newval[1];
				(*it)[idx+2] = newval[2];

				// reset all previous counters
				for (boost::dynamic_bitset<>::size_type i = 0; i < idx; i++) {
					(*it)[i] = 0;
				}

				break;
			}
		}
	}

	/**
	 * Returns the position of the split in the base_split array
	 * for a specific edge identified by its index position, according to the current
	 * state of the split.
	 *
	 * @param idx the index position of the edge in the edges list
	 * @return the position of the split in the base_split array
	 */
	inline short int get_split_pos(boost::dynamic_bitset<>::size_type idx) {
		bit_set mask = bit_set(it->size());
		mask.set(0); mask.set(1); mask.set(2);
		idx *= 3;
		return (((mask <<= idx) & (*it)) >>= idx).to_ulong() + 1;
	}

	inline RELTYPE get_split_for_relation(boost::dynamic_bitset<>::size_type idx) {
		std::list<RELTYPE>::const_iterator itl;
		unsigned int counter = 0;
		for (itl = rels.begin(); itl != rels.end(); ++itl, counter++) {
			if (counter == idx) {
				return base_split[*itl][get_split_pos(idx)];
			}
		}

		return 0;
	}

	inline long double get_max_value() const {
		long double combs = 1;
		std::list<RELTYPE>::const_iterator itl;
		for (itl = rels.begin(); itl != rels.end(); ++itl) {
			combs *= base_split[*itl][0];
		}

		return combs;
	}
} bsplit_set;


/**
 * Creates a split_set if it is not already created, and then updates it by adding
 * an edge in case it can be split in horn relations.
 *
 * @param split the split set to create or update
 * @param r the RCC-8 relation of the edge specified in \a edge
 * @param edge the edge with which the split set in \a split will be updated
 * @param three_split true when the inserted edge can be split in three horn relations
 * @return the updated split or NULL if the split set in \a split was NULL and no edge
 *         was inserted
 */
split_set* create_or_update_split(split_set *split, RELTYPE r, edge_t edge, bool *three_split);

bsplit_set* create_or_update_base_split(bsplit_set *bsplit, RELTYPE r, edge_t edge);

/**
 * Compares two RCC-8 constraints according to their restrictiveness of their
 * neighborhoods and the weighting scheme used in [van Beek/Manchak, 1996].
 *
 * If the restrictiveness of two RCC-8 constraints are the same, then we consider
 * more restrictive the one with fewer base RCC-8 relations. For example,
 * according to [van Beek/Manchak, 1996], the restrictiveness value for the
 * relation {EC, NTPP} is 5, while for {TPP, TPPI, EQ} is 5 as well. However,
 * between the two, we regard the first one as more restrictive to return for
 * splitting, since it would catch a quicker inconsistency.
 *
 * @param i first RCC-8 relation
 * @param j second RCC-8 relation
 * @return true when the first argument \a i is more restrictive (goes before
 * 		   the second argument in the order of restrictiveness) than the second
 * 		   argument \a j; false otherwise
 */
bool compare_constraints_vb(const rcc8_con &i, const rcc8_con &j);

inline int compute_weight_vb(RELTYPE rel) {
	int weight = 0;

	// compute vanBeek weights
	for (int m = 0, mask = 1; m < MAXREL; m++, mask = 1 << m) {
		int lwght = vb_weight[m];

		if (rel & mask) {
			weight += lwght;
		}
	}

	return weight;
}

inline bsplit_set *create_atomic_bsplit_set(const edge_t &edge, RELTYPE rel) {
	bsplit_set *split = new bsplit_set();

	split->it = new bit_set(3);
	split->edges.push_back(edge);
	split->rels.push_back(rel);

	return split;
}

inline split_set *create_atomic_horn_split_set(const edge_t &edge, RELTYPE rel) {
	split_set *split = new split_set();

	bool three_split_sz = (horn_split[rel][0] == 3) ? 1:0;

	split->it = new bit_set(1);
	split->msb_it = new bit_set(three_split_sz);
	split->edges.push_back(edge);

	return split;
}

short int num_base_rels(RELTYPE r);

std::string print_relation(RELTYPE r);

inline short int get_num_hsplits(RELTYPE r) {
	return horn_split[r][0];
}

inline short int get_num_bsplits(RELTYPE r) {
	return base_split[r][0];
}

#endif /* RCC8_H_ */
