/**
 * This Source Code Form is subject to the terms of the Mozilla Public
 * License, v. 2.0. If a copy of the MPL was not distributed with this
 * file, You can obtain one at http://mozilla.org/MPL/2.0/.
 *
 * Copyright (C) 2013, 2014 Department of Informatics and
 * Telecommunications, National and Kapodistrian University of Athens
 *
 * Contains modified code from the RCC-8 reasoner implemented by
 * Ronny Fehling, Bernhard Nebel, and Jochen Renz at the Department of
 * Informatics, University of Freiburg, Germany.
 *
 *  Created on: Dec 19, 2013
 *      Author: Charalampos Nikolaou <charnik@di.uoa.gr>
 */
#include "rcc8csp.h"
#include <iostream>

// define static members (composition table and table of inverse relations)
RELTYPE RCC8_CSP::fcomp[MAXSET][MAXSET];
RELTYPE RCC8_CSP::invtab[MAXSET];

void RCC8_CSP::init() {
	csp = NULL;
	mark = NULL;
	fixed = NULL;
	split_it = NULL;
	horn_csp = NULL;
	nnodes = 0;

	pcits = 0;
	pcops = 0;
}

RCC8_CSP::RCC8_CSP(int nodes) {
	init();
	nnodes = nodes;

	create_matrices(NULL, NULL);
}

RCC8_CSP::RCC8_CSP(int nodes, const RELTYPE *matrix, const char *fixed_matrix, const nodeid_map &mapping) {
	init();
	nnodes = nodes;
	create_matrices(matrix, fixed_matrix);
	nodeids = mapping;
}

RCC8_CSP::RCC8_CSP() {
	init();
}

RCC8_CSP::RCC8_CSP(const std::set<idx_t> &node_ids,
				   const edge_set &edges,
				   const idx_t *xadj, const idx_t *adjncy, const RELTYPE *adjwgt) {
	init();
	nnodes = node_ids.size();

	// map node identifiers to internal node identifiers (consecutive starting from zero)
	std::set<idx_t>::const_iterator its;
	int id = 0;
	for (its = node_ids.begin(); its != node_ids.end(); ++its) {
		nodeids[*its] = id++;
		//std::cout << *its << " --> " << nodeids.at(*its) << "\n";
	}
	assert (id == nnodes);

	// initialize matrices
	create_matrices(NULL, NULL);

	// populate matrices
	edge_set::const_iterator itl;
	for (itl = edges.begin(); itl != edges.end(); ++itl) {
		idx_t real_src_id = itl->first;
		idx_t real_trgt_id = itl->second;

//		assert (real_src_id < real_trgt_id);

		// get internal ids
		idx_t src_internal = nodeids.at(real_src_id);
		idx_t trgt_internal = nodeids.at(real_trgt_id);

		if (src_internal > trgt_internal) { // swap, we need upper triangular
			idx_t temp = src_internal;
			src_internal = trgt_internal;
			trgt_internal = temp;

			// we should swap the real ids as well
			real_src_id = itl->second;
			real_trgt_id = itl->first;
		}

		// find index of edge in CSR format
		idx_t idx = Util::csr_edge_index(real_src_id, real_trgt_id, xadj, adjncy);
		assert (idx > -1);

//		assert (adjwgt[idx] == RCC8_CSP::inverse(adjwgt[Util::csr_edge_index(real_trgt_id, real_src_id, xadj, adjncy)]));

		// set the edge label
		csp[TRINDX(src_internal, trgt_internal)] = adjwgt[idx];
	}
}

bool RCC8_CSP::create_matrices(const RELTYPE *toclone, const char *fixed_matrix) {
	long onedim_sz = RCC8_CSP::matrix_size(nnodes);

	if ((csp = new RELTYPE[onedim_sz]) == NULL) {
		perror("Cannot allocate network's matrix.");
		csp = NULL;
		return false;
	}

	if (toclone == NULL) {
		// initialize network's matrix (upper triangular)
		std::fill_n(csp, onedim_sz, DALL);

	} else { // clone the given network
		memcpy(csp, toclone, onedim_sz * sizeof(RELTYPE));
	}

	if ((mark = new char[onedim_sz]) == NULL) {
		perror("Cannot allocate Renz's mark matrix.");
		FREE_ARRAY(csp);
		return false;
	}

	if ((fixed = new char[onedim_sz]) == NULL) {
		perror("Cannot allocate fixed matrix.");
		FREE_ARRAY(mark);
		FREE_ARRAY(csp);
		return false;
	}

	if (fixed_matrix == NULL) {
		// initialize fixed matrix
		std::fill_n(fixed, onedim_sz, 0);

	} else {
		memcpy(fixed, fixed_matrix, onedim_sz * sizeof(char));
	}

	return true;
}

bool RCC8_CSP::consistent() {
	if (!path_consistent()) return false;

	if (false) { // condition to switch to the iterative version using refinement iterators
		// restore or compute split set
		split_set* split = (split_it != NULL) ? split_it:create_split_set();

		// the network is path-consistent and contains relations from a tractable set
		if (split == NULL) return true;

		while (!split->end()) { // while the split iterator is not at the end
			// create a copy of the current RCC-8 network according to the split iterator
			RCC8_CSP *newcsp = clone();
			newcsp->apply_local_split(split);

			if (newcsp->path_consistent()) { // path consistent
				// save the position of the iterator as well as the path-consistent horn RCC-8 network
				split_it = split;

				// make sure you free up space of the previous path-consistent horn RCC-8 network
				FREE_PTR(horn_csp);
				horn_csp = newcsp;

				// persist the path-consistent network
				//set_csp_network(horn_csp->get_csp_network());

				return true;
			}

			// clean and check next split
			FREE_PTR(newcsp);

			// increment iterator
			split->incr();
		}

		return false;

	} else { // use the traditional backtracking algorithm
		// get the next split
		split_set *split = compute_next_split();

		// no other split to make
		if (split == NULL) return true;

		bool cons_result = false;
		while (!split->end() && !cons_result) {
			// create a copy of the current RCC-8 network
			RCC8_CSP *newcsp = clone();

			if (newcsp == NULL) {
				std::cerr << "RCC8_CSP: out of memory (while backtracking)\n";
				break;
			}

			newcsp->apply_local_split(split);

			if (newcsp->consistent()) {
				cons_result = true;

			} else { // try the next refinement
				// release resources
				FREE_PTR(newcsp);

				split->incr();
			}
		}

		// release split
		FREE_SPLIT(split);

		return cons_result;
	}
}

split_set *RCC8_CSP::compute_next_split() const {
	std::list<rcc8_con> constraints;

	for (int i = 0; i < nnodes; i++) {
		for (int j = i + 1; j < nnodes; j++) {
			long tridx = TRINDX(i, j);

			if (!fixed[tridx]) { // edge not belonging to the set of shared edges
				RELTYPE rel = csp[tridx];

				// if the edge can be split in more than one horn relations and
				if (horn_split[rel][0] > 1) {
					constraints.push_back(boost::make_tuple(i, j, rel));
				}
			}
		}
	}

	split_set* split = NULL;
	if (!constraints.empty()) {
		// sort them according to their restrictiveness (first would be the more restrictive one)
		constraints.sort(compare_constraints_vb);

		idx_t src = constraints.begin()->get<0>();
		idx_t trgt = constraints.begin()->get<1>();
		RELTYPE tosplit = constraints.begin()->get<2>();

		split  = create_atomic_horn_split_set(std::make_pair(src, trgt), tosplit);
	}

	return split;
}

void RCC8_CSP::apply_local_split(split_set *split) {
	edge_list::const_iterator itl;
	boost::dynamic_bitset<>::size_type idx = split->edges.size() - 1;
	for (itl = split->edges.begin(); itl != split->edges.end(); ++itl, --idx) {
		long edge_idx = TRINDX(itl->first, itl->second);
		RELTYPE rel_to_split = csp[edge_idx];

//			std::cout << print_rel(rel_to_split) << " --> ";

		csp[edge_idx] = horn_split[rel_to_split][split->get_split_pos(idx)];

//			std::cout << print_rel(csp[edge_idx]) << "\n";
	}
}

bool RCC8_CSP::reset_split() {
	if (split_it != NULL) {
		// reset two-split counter
		if (split_it->it->size() < split_it->edges.size()) { // numbering of three-splits has been started
			delete split_it->it;
			split_it->it = new bit_set(split_it->edges.size());

		} else { // just clear all bits
			split_it->it->reset();

		}

		// reset three-split counter
		split_it->msb_it->reset();

		return true;
	}

	return false;
}

bool RCC8_CSP::apply_global_split(split_set *split) {
	edge_list::const_iterator itl;
	boost::dynamic_bitset<>::size_type idx = split->edges.size() - 1;
	for (itl = split->edges.begin(); itl != split->edges.end(); ++itl, --idx) {

		// get internal ids
		nodeid_map::iterator src_it = nodeids.find(itl->first);
		if (src_it != nodeids.end()) {
			nodeid_map::iterator trgt_it = nodeids.find(itl->second);

			if (trgt_it != nodeids.end()) {
				idx_t src_internal = src_it->second;
				idx_t trgt_internal = trgt_it->second;

				if (src_internal > trgt_internal) { // take this edge for the upper triangular
					idx_t swap = src_internal;
					src_internal = trgt_internal;
					trgt_internal = swap;
				}
// FIXME like the one for base split
				long edge_idx = TRINDX(src_internal, trgt_internal);
				RELTYPE rel_to_split = csp[edge_idx];
				csp[edge_idx] = horn_split[rel_to_split][split->get_split_pos(idx)];

				// mark edge as not modifiable
				fixed[edge_idx] = 1;
			}
		}
	}

	// FIXME like the one for base split
	return true;
}

bool RCC8_CSP::apply_global_split(bsplit_set *split) {
	bool changed = false;
	edge_list::const_iterator itl;
	std::list<RELTYPE>::const_iterator itrel;

	boost::dynamic_bitset<>::size_type idx = split->edges.size() - 1;
	for (itl = split->edges.begin(), itrel = split->rels.begin(); itl != split->edges.end(); ++itl, ++itrel, --idx) {

		// get internal ids
		nodeid_map::iterator src_it = nodeids.find(itl->first);
		if (src_it != nodeids.end()) {
			nodeid_map::iterator trgt_it = nodeids.find(itl->second);

			if (trgt_it != nodeids.end()) {
				idx_t src_internal = src_it->second;
				idx_t trgt_internal = trgt_it->second;

				ENSURE_ORDER(src_internal, trgt_internal);

				long edge_idx = TRINDX(src_internal, trgt_internal);
//				std::cout << print_rel(csp[edge_idx]) << " --> ";

				csp[edge_idx] = base_split[*itrel][split->get_split_pos(idx)];

//				std::cout << print_rel(csp[edge_idx]) << "\n";

				// mark edge as not modifiable
				fixed[edge_idx] = 1;
				changed = true;
			}
		}
	}

	return changed;
}

split_set* RCC8_CSP::create_split_set() {
	split_set* split = NULL;

	int split_in_three = 0;
	for (int i = 0; i < nnodes; i++) {
		for (int j = i + 1; j < nnodes; j++) {
			long tridx = TRINDX(i, j);

			if (!fixed[tridx]) { // edge not belonging to the set of shared edges
				int cnt = horn_split[csp[tridx]][0];

				// if the edge can be split in more than one horn relations and
				if (cnt > 1) {
					if (split == NULL) { // create once
						split = new split_set();
					}

					// keep the edge
					if (cnt == 2) { // split of size two (put it back)
						split->edges.push_back(std::make_pair(i, j));

					} else { // split of size three (put it in the front)
						split->edges.push_front(std::make_pair(i, j));
						split_in_three++;
					}
				}
			}
		}
	}

	if (split != NULL) { // we have found some edges, initialize the rest data structure
		split->it = new bit_set(split->edges.size());
		split->msb_it = new bit_set(split_in_three);
	}

	return split;
}

bool RCC8_CSP::is_horn() {
	for (int i = 0; i < nnodes; i++) {
		for (int j = i + 1; j < nnodes; j++) {
			// if the edge can be split in more than one horn relations
			if (horn_split[csp[TRINDX(i, j)]][0] > 1) {
				return false;
			}
		}
	}

	return true;
}

// TODO optimize to compute index positions only once
//bool enable_fixed
bool RCC8_CSP::path_consistent() {
	long node1 = -1, node2 = -1; /* If node1, node2 > -1 then use only this as initial "queue" */
	register long i, j, k;
	register RELTYPE temp;
	bool changed;

	pcits = 0;
	pcops = 0;

	if (node1 < 0) {
		for (i = 0; i < nnodes; i++) {
			for (j = i + 1; j < nnodes; j++) {
				//mark[i][j] = 1;
				mark[TRINDX(i, j)] = 1;
			}
		}

	} else {
		for (i = 0; i < nnodes; i++) {
			for (j = i + 1; j < nnodes; j++) {
				//mark[i][j] = 0;
				mark[TRINDX(i, j)] = 0;
			}
		}

		//mark[node1][node2] = 1;
		mark[TRINDX(node1, node2)] = 1;
	}

	do {
		changed = false;
		pcits++;
		for (i = 0; i < nnodes; i++)
			for (j = i + 1; j < nnodes; j++)
				//if (mark[i][j]) {
				if (mark[TRINDX(i, j)]) {
					//mark[i][j] = 0;
					mark[TRINDX(i, j)] = 0;

					for (k = 0; k < nnodes; k++) {
						if ((k != i) && (k != j)) {
							pcops++; /* constrain k,i,j */
							if (k < i) {
								//temp = FCOMP1(csp[k][i],csp[i][j]);
								temp = FCOMP1(csp[TRINDX(k, i)], csp[TRINDX(i, j)]);

							} else {
								//temp = FCOMP1(FINV(csp[i][k]),csp[i][j]);
								temp = FCOMP1(FINV(csp[TRINDX(i, k)]), csp[TRINDX(i, j)]);

							}

							if (temp != DALL) {
								if (k < j) {
									//temp = (temp&csp[k][j]);*/
									temp = (temp & csp[TRINDX(k, j)]);

									if (temp == 0) {
										return false;
									}

									//if (temp != csp[k][j]) {
									if (temp != csp[TRINDX(k, j)]) {
										changed = true;

										// check change of hard constraints
										//if (enable_fixed && fixed[TRINDX(k, j)]) return false;

											//mark[k][j] = 1;*/
										mark[TRINDX(k, j)] = 1;

										//csp[k][j] = temp;
										csp[TRINDX(k, j)] = temp;

									}
								} else {
									temp = (FINV(temp) & csp[TRINDX(j, k)]);
									/*temp = (FINV(temp)&csp[j][k]);*/
									if (temp == 0) {
										return false;
									}

									if (temp != csp[TRINDX(j, k)]) {
										/*if (temp != csp[j][k]) {*/
										changed = true;

										// check change of hard constraints
										//if (enable_fixed && fixed[TRINDX(j, k)]) return false;

										mark[TRINDX(j, k)] = 1;
										/*mark[j][k] = 1;*/

										csp[TRINDX(j, k)] = temp;
										/*csp[j][k] = temp;*/
									}
								}
							}

							pcops++; /* constrain i,j,k */
							if (j < k) {
								//temp = FCOMP1(csp[i][j],csp[j][k]);
								temp = FCOMP1(csp[TRINDX(i, j)], csp[TRINDX(j, k)]);

							} else {
								//temp = FCOMP1(csp[i][j],FINV(csp[k][j]));
								temp = FCOMP1(csp[TRINDX(i, j)], FINV(csp[TRINDX(k, j)]));

							}
							if (temp != DALL) {
								if (i < k) {
									//temp = (temp&csp[i][k]);
									temp = (temp & csp[TRINDX(i, k)]);

									if (temp == 0) {
										return false;
									}

									/*if (temp != csp[i][k]) {*/
									if (temp != csp[TRINDX(i, k)]) {
										changed = true;

										// check change of hard constraints
										//if (enable_fixed && fixed[TRINDX(i, k)]) return false;

										//mark[i][k] = 1;*/
										mark[TRINDX(i, k)] = 1;

										//csp[i][k] = temp;
										csp[TRINDX(i, k)] = temp;
									}

								} else {
									//temp = (FINV(temp)&(csp[k][i]));
									temp = (FINV(temp) & (csp[TRINDX(k, i)]));

									if (temp == 0) {
										return false;
									}

									//if (temp != csp[k][i]) {
									if (temp != csp[TRINDX(k, i)]) {
										changed = true;

										// check change of hard constraints
										//if (enable_fixed && fixed[TRINDX(k, i)]) return false;

										//mark[k][i] = 1;
										mark[TRINDX(k, i)] = 1;

										//csp[k][i] = temp;
										csp[TRINDX(k, i)] = temp;
									}
								}
							}
						}
					}
				}
	} while (changed);

	return true;
}

RELTYPE RCC8_CSP::inverse(RELTYPE r) {
	return( ((r&DC) 	 ? DC 	: 0) |
			((r&EC) 	 ? EC 	: 0) |
			((r&PO) 	 ? PO 	: 0) |
			((r&TPP)   ? TPPI : 0) |
			((r&NTPP)  ? NTPPI: 0) |
			((r&TPPI)  ? TPP 	: 0) |
			((r&NTPPI) ? NTPP : 0) |
			((r&EQ) 	 ? EQ  	: 0) );
}


int RCC8_CSP::relvalue(RELTYPE r) {
  register int bytecnt;
  register int tempval;

  bytecnt = 0;
  tempval = 0;
  if (r & DC) {
    bytecnt++;
    tempval += 6;
  }
  if (r & EC) {
    bytecnt++;
    tempval += 3;
  }
  if (r & PO) {
    bytecnt++;
    tempval += 5;
  }
  if (r & TPP) {
    bytecnt++;
    tempval += 4;
  }
  if (r & NTPP) {
    bytecnt++;
    tempval += 2;
  }
  if (r & TPPI) {
    bytecnt++;
    tempval += 4;
  }
  if (r & NTPPI) {
    bytecnt++;
    tempval += 2;
  }
  if (r & EQ) {
    bytecnt++;
    tempval += 1;
  }
  return(tempval);
}

void RCC8_CSP::compute_comp_table() {
  register RELTYPE hibit=2,i,j;

  for(j=3;j<MAXSET;j++)
    if (j % hibit)
      for(i=1;i<MAXSET;i*=2)
	RCC8_CSP::fcomp[i][j] = (RCC8_CSP::fcomp[i][hibit])|(RCC8_CSP::fcomp[i][j^hibit]);
    else
      hibit*=2;

  hibit=2;
  for(i=3;i<MAXSET;i++)
    if (i % hibit)
      for(j=1;j<MAXSET;j++)
    RCC8_CSP::fcomp[i][j] = (RCC8_CSP::fcomp[hibit][j])|(RCC8_CSP::fcomp[i^hibit][j]);
    else
      hibit*=2;
}

void RCC8_CSP::initfop() {
  register RELTYPE r1;

  /* Basic composition table */

  /*DC*/
  RCC8_CSP::fcomp[DC][DC] = DALL;
  RCC8_CSP::fcomp[DC][EC] = (DC|EC|PO|TPP|NTPP);
  RCC8_CSP::fcomp[DC][PO] = (DC|EC|PO|TPP|NTPP);
  RCC8_CSP::fcomp[DC][TPP] = (DC|EC|PO|TPP|NTPP);
  RCC8_CSP::fcomp[DC][NTPP] = (DC|EC|PO|TPP|NTPP);
  RCC8_CSP::fcomp[DC][TPPI] = DC;
  RCC8_CSP::fcomp[DC][NTPPI] = DC;
  RCC8_CSP::fcomp[DC][EQ] = DC;

  /*EC*/
  RCC8_CSP::fcomp[EC][DC] = (DC|EC|PO|TPPI|NTPPI);
  RCC8_CSP::fcomp[EC][EC] = (DC|EC|PO|TPP|TPPI|EQ);
  RCC8_CSP::fcomp[EC][PO] = (DC|EC|PO|TPP|NTPP);
  RCC8_CSP::fcomp[EC][TPP] = (EC|PO|TPP|NTPP);
  RCC8_CSP::fcomp[EC][NTPP] = (PO|TPP|NTPP);
  RCC8_CSP::fcomp[EC][TPPI] = (DC|EC);
  RCC8_CSP::fcomp[EC][NTPPI] = DC;
  RCC8_CSP::fcomp[EC][EQ] = EC;

  /*PO*/
  RCC8_CSP::fcomp[PO][DC] = (DC|EC|PO|TPPI|NTPPI);
  RCC8_CSP::fcomp[PO][EC] = (DC|EC|PO|TPPI|NTPPI);
  RCC8_CSP::fcomp[PO][PO] = DALL;
  RCC8_CSP::fcomp[PO][TPP] = (PO|TPP|NTPP);
  RCC8_CSP::fcomp[PO][NTPP] = (PO|TPP|NTPP);
  RCC8_CSP::fcomp[PO][TPPI] = (DC|EC|PO|TPPI|NTPPI);
  RCC8_CSP::fcomp[PO][NTPPI] = (DC|EC|PO|TPPI|NTPPI);
  RCC8_CSP::fcomp[PO][EQ] = PO;

  /*TPP*/
  RCC8_CSP::fcomp[TPP][DC] = DC;
  RCC8_CSP::fcomp[TPP][EC] = (DC|EC);
  RCC8_CSP::fcomp[TPP][PO] = (DC|EC|PO|TPP|NTPP);
  RCC8_CSP::fcomp[TPP][TPP] = (TPP|NTPP);
  RCC8_CSP::fcomp[TPP][NTPP] = NTPP;
  RCC8_CSP::fcomp[TPP][TPPI] = (DC|EC|PO|TPP|TPPI|EQ);
  RCC8_CSP::fcomp[TPP][NTPPI] = (DC|EC|PO|TPPI|NTPPI);
  RCC8_CSP::fcomp[TPP][EQ] = TPP;

  /*NTPP*/
  RCC8_CSP::fcomp[NTPP][DC] = DC;
  RCC8_CSP::fcomp[NTPP][EC] = DC;
  RCC8_CSP::fcomp[NTPP][PO] = (DC|EC|PO|TPP|NTPP);
  RCC8_CSP::fcomp[NTPP][TPP] = NTPP;
  RCC8_CSP::fcomp[NTPP][NTPP] = NTPP;
  RCC8_CSP::fcomp[NTPP][TPPI] = (DC|EC|PO|TPP|NTPP);
  RCC8_CSP::fcomp[NTPP][NTPPI] = DALL;
  RCC8_CSP::fcomp[NTPP][EQ] = NTPP;

  /*TPPI*/
  RCC8_CSP::fcomp[TPPI][DC] = (DC|EC|PO|TPPI|NTPPI);
  RCC8_CSP::fcomp[TPPI][EC] = (EC|PO|TPPI|NTPPI);
  RCC8_CSP::fcomp[TPPI][PO] = (PO|TPPI|NTPPI);
  RCC8_CSP::fcomp[TPPI][TPP] = (PO|TPP|TPPI|EQ);
  RCC8_CSP::fcomp[TPPI][NTPP] = (PO|TPP|NTPP);
  RCC8_CSP::fcomp[TPPI][TPPI] = (TPPI|NTPPI);
  RCC8_CSP::fcomp[TPPI][NTPPI] = NTPPI;
  RCC8_CSP::fcomp[TPPI][EQ] = TPPI;

  /*NTPPI*/
  RCC8_CSP::fcomp[NTPPI][DC] = (DC|EC|PO|TPPI|NTPPI);
  RCC8_CSP::fcomp[NTPPI][EC] = (PO|TPPI|NTPPI);
  RCC8_CSP::fcomp[NTPPI][PO] = (PO|TPPI|NTPPI);
  RCC8_CSP::fcomp[NTPPI][TPP] = (PO|TPPI|NTPPI);
  RCC8_CSP::fcomp[NTPPI][NTPP] = (PO|TPP|TPPI|NTPP|NTPPI|EQ);
  RCC8_CSP::fcomp[NTPPI][TPPI] = NTPPI;
  RCC8_CSP::fcomp[NTPPI][NTPPI] = NTPPI;
  RCC8_CSP::fcomp[NTPPI][EQ] = NTPPI;

  /*EQ*/
  RCC8_CSP::fcomp[EQ][DC] = DC;
  RCC8_CSP::fcomp[EQ][EC] = EC;
  RCC8_CSP::fcomp[EQ][PO] = PO;
  RCC8_CSP::fcomp[EQ][TPP] = TPP;
  RCC8_CSP::fcomp[EQ][NTPP] = NTPP;
  RCC8_CSP::fcomp[EQ][TPPI] = TPPI;
  RCC8_CSP::fcomp[EQ][NTPPI] = NTPPI;
  RCC8_CSP::fcomp[EQ][EQ] = EQ;


  compute_comp_table();

  for (r1 = 0; r1 < MAXSET; r1++) {
	  RCC8_CSP::invtab[r1] = inverse(r1);
  }
}

// FIXME this method hasn't been finished yet!
bool RCC8_CSP::read(const char* filename, idx_t *sz, idx_t **xadj, idx_t **ajncy, RELTYPE **adjwgt) {
	std::string line;
	int i, j;
	RELTYPE r;
	std::string csp_name;	/* the description name of the CSP (ignored for the time being) */

	std::ifstream cnet(filename);
	if (!cnet.is_open()) {
		std::cerr << "RCC8_CSP: error while opening CSP file " << filename << std::endl;
		return false;
	}

	do { // skip empty lines
		std::getline(cnet, line);
	} while (line.size() == 0 && !cnet.eof());

	if (cnet.eof()) {
		return false;
	}

	// read header line
	std::stringstream strm(line);

	strm >> *sz; // get maximum node id
	strm >> csp_name; // get description

	// debug
	//std::cout << *sz << ": " << csp_name << std::endl;

	(*sz)++;

	do {
		std::getline(cnet, line);

		if (line[0] == '.') { // finish reading CSP network
			return true;
		}

		if (parsecons(line.c_str(), &r, &i, &j)) {
			if ((i > *sz - 1) || (i < 0) || (j > *sz - 1) || (j < 0)
					|| (i == j)) {
				std::cerr << "RCC8_CSP: illegal node id: '" << line << "'\n";
				return false;
			}

			if (i > j) {
				// TODO populate arrays
				//csp[j][i] &= inverse(r);

			} else {
				// TODO populate arrays
				//csp[i][j] &= r;
				//std::string temp;
				//std::cout << "(" << i  << ", " << j << ") " << RCC8_CSP::print_rel(temp, r) << "\n";

			}
		} else {
			return false;
		}
	} while (true);

	return true; // will never hit
}

/* Parse relation from string */
int parserel(char *line, RELTYPE *r)
{
  RELTYPE result;
  char *oldline;

  oldline = line;
  result = 0;


  while ((*line != '\0') && (*line != '(')) line++;
  if (*(line++) != '(') {
    fprintf(stderr,"\n*** no '(' in rel spec: '%s'\n",oldline);
    return(0);
  }
  while (*line == ' ') line++;
  while ((*line != ')' ) && *line) {
    while (*line == ' ') line++;
    switch (*(line++)) {
    case 'D' : case 'd' :
      result |= DC;
      line++;
      break;
    case 'P' : case 'p' :
      result |= PO;
      line++;
      break;
    case 'E' : case 'e' :
      if (*line == 'C' || *line == 'c') result |= EC;
      else result |= EQ;
      line++;
      break;
    case 'N' : case 'n':
      if (*(line=line+3) == 'I' || *line == 'i') result |= NTPPI;
      else result |= NTPP;
      line++;
      break;
    case 'T' : case 't':
      if (*(line=line+2) == 'I' || *line == 'i') result |= TPPI;
      else result |= TPP;
      line++;
      break;
    default :
      fprintf(stderr,"\n*** Illegal rel spec: '%s'\n",oldline);
      return(0);
    }
    while (*line == ' ') line++;
  }
  if (*line != ')') {
    fprintf(stderr,"\n*** Missing ')' in rel spec in '%s'\n",oldline);
    return(0);
  }
  *r = result;
  return(1);
}



/* Parse constraint from string */
int RCC8_CSP::parsecons(const char *line, RELTYPE *r, int *node1, int *node2) {
	*node1 = -1;
	*node2 = -1;

	if (sscanf(line, "%d %d", node1, node2) != 2) {
		std::cerr << "RCC8_CSP: no node ids in rel spec: '" << line << "'\n";
		return(0);
	}

  return(RCC8_CSP::parserel(line, r));
}

/* Parse relation from string */
int RCC8_CSP::parserel(const char *line, RELTYPE *r) {
  RELTYPE result;
  const char *oldline = line;

  result = 0;

  while ((*line != '\0') && (*line != '(')) line++;
  if (*(line++) != '(') {
    fprintf(stderr,"\n*** no '(' in rel spec: '%s'\n",oldline);
    return(0);
  }
  while (*line == ' ') line++;
  while ((*line != ')' ) && *line) {
    while (*line == ' ') line++;
    switch (*(line++)) {
    case 'D' : case 'd' :
      result |= DC;
      line++;
      break;
    case 'P' : case 'p' :
      result |= PO;
      line++;
      break;
    case 'E' : case 'e' :
      if (*line == 'C' || *line == 'c') result |= EC;
      else result |= EQ;
      line++;
      break;
    case 'N' : case 'n':
      if (*(line=line+3) == 'I' || *line == 'i') result |= NTPPI;
      else result |= NTPP;
      line++;
      break;
    case 'T' : case 't':
      if (*(line=line+2) == 'I' || *line == 'i') result |= TPPI;
      else result |= TPP;
      line++;
      break;
    default :
      fprintf(stderr,"\n*** Illegal rel spec: '%s'\n",oldline);
      return(0);
    }
    while (*line == ' ') line++;
  }
  if (*line != ')') {
    fprintf(stderr,"\n*** Missing ')' in rel spec in '%s'\n",oldline);
    return(0);
  }
  *r = result;
  return(1);
}

/* Print relation to string */
std::string RCC8_CSP::print_rel(RELTYPE r) {
  return print_relation(r);
}

int RCC8_CSP::check_and_refine(RCC8_CSP *cspl, RCC8_CSP *cspr, const edge_list &shared_edges) {
	int refined = UNCHANGED;

	edge_list::const_iterator it;
	for (it = shared_edges.begin(); it != shared_edges.end(); ++it) {
		// get internal ids for left CSP
		idx_t l_i = cspl->nodeids.at(it->first);
		idx_t l_j = cspl->nodeids.at(it->second);

		// swap, we need upper triangular
		ENSURE_ORDER(l_i, l_j);

		// get internal ids for right CSP
		idx_t r_i = cspr->nodeids.at(it->first);
		idx_t r_j = cspr->nodeids.at(it->second);

		// swap, we need upper triangular
		ENSURE_ORDER(r_i, r_j);

		// compute edge indices
		idx_t le_idx = TRINDX(l_i, l_j);
		idx_t re_idx = TRINDX(r_i, r_j);

		// get respective RCC-8 relations
		RELTYPE lrel = cspl->csp[le_idx];
		RELTYPE rrel = cspr->csp[re_idx];

		if (lrel == rrel) { // equal
			continue; // continue to the next edge

		} else {
			// get the intersection
			RELTYPE new_rel = lrel & rrel;
			if (new_rel == 0) { // inconsistent
				// mark inconsistency
				refined = INCONSISTENT;

				// keep inconsistency (just for provenance)
				cspl->csp[le_idx] = new_rel;
				cspr->csp[re_idx] = new_rel;

				break;	// do not need to check other edges

			} else {
				if ((new_rel != lrel) && (new_rel != rrel)) {
					cspl->csp[le_idx] = cspr->csp[re_idx] = new_rel;
					refined |= REFINED_BOTH;

				} else if (new_rel != lrel) {
					cspl->csp[le_idx] = new_rel;
					refined |= REFINED_LEFT;

				} else if (new_rel != rrel) {
					cspr->csp[re_idx] = new_rel;
					refined |= REFINED_RIGHT;

				}
			}
		}
	}

	return refined;
}
