/**
 * This Source Code Form is subject to the terms of the Mozilla Public
 * License, v. 2.0. If a copy of the MPL was not distributed with this
 * file, You can obtain one at http://mozilla.org/MPL/2.0/.
 *
 * Copyright (C) 2013, 2014 Department of Informatics and
 * Telecommunications, National and Kapodistrian University of Athens
 *
 * Contains modified code from the RCC-8 reasoner implemented by
 * Ronny Fehling, Bernhard Nebel, and Jochen Renz at the Department of
 * Informatics, University of Freiburg, Germany.
 *
 *  Created on: Dec 19, 2013
 *      Author: Charalampos Nikolaou <charnik@di.uoa.gr>
 */
#ifndef RCC8CSP_H_
#define RCC8CSP_H_

#include <cstdio>
#include <cstring>
#include <fstream>
#include <sstream>
#include <set>
#include <list>
#include <vector>
#include <utility>
#include <cassert>

#include <boost/unordered_map.hpp>

#include "metis/metis.h"
#include "Util.h"
#include "rcc8.h"

#define FINV(r)			(RCC8_CSP::invtab[r])
//#define FVAL(r)			(vtab[r])

/* macro for accessing the composition table */
#define FCOMP1(rel1,rel2) RCC8_CSP::fcomp[rel1][rel2]

/* table holding the splits in horn relations for an RCC-8 relation */
extern RELTYPE horn_split[MAXSET][4];

/* Whoever needs access to triangulated matrices should use
 * the following macro that abstract their type.
 */
#define TRINDX(i, j)	(i) + (((j) * ((j) + 1)) / 2)

#define ENSURE_ORDER(src, trgt)	if ((src) > (trgt)) {idx_t _temp = (src); (src) = (trgt); (trgt) = _temp;}

/* map a nodeid to an internal auto-increment id (for used in matrice) */
typedef boost::unordered_map<idx_t, idx_t> nodeid_map;

// for their use, see static method RCC8_CSP::check_and_refine
#define INCONSISTENT	-1
#define UNCHANGED		 0
#define REFINED_RIGHT	 1
#define REFINED_LEFT	 2
#define REFINED_BOTH	 3

class ConsistencyEngine;

class RCC8_CSP {
	friend class ConsistencyEngine;

private:
	RELTYPE *csp;			/**< triangulated matrix holding the RCC-8 network */
	char *mark;				/**< triangulated mark matrix */
	char *fixed;			/**< triangulated fixed matrix (edges that should not change) */
	int nnodes;				/**< number of nodes of the RCC-8 network */

	nodeid_map nodeids;		/**< a map associating scattered node ids to auto-increment ids
								 to be used by the underlying constraint network using the
								 internal triangular matrix format. This map is used when the
								 network is initialized by the partition graph, \a PartGraph */
	split_set* split_it;	/**< if it is not NULL, then it holds the position of the iterator
	  	  	  	  	  	  	  	 over all combinations of split edges into horn relations */
	RCC8_CSP *horn_csp;		/**< if it is not NULL, then it holds the RCC-8 constraint network
								 which is composed only of horn relations and is path-consistent */

	// statistics
	long pcits;				/**< number of iterations for path-consistency */
	long pcops;				/**< number of path-consistency operations */

	/**
	 * Allocates memory for the internal matrix data structures. If another
	 * matrix is given as an argument, it clones it.
	 *
	 * @param toclone [optional] a matrix to be copied
	 * @return true in case of success, false otherwise.
	 */
	bool create_matrices(const RELTYPE *csp_matrix, const char *fixed_matrix);

	/**
	 * Initializes internal state.
	 */
	void init();

//	/**
//	 * Computes the internal node ids for a given edge with global node ids.
//	 *
//	 * @param src the internal local id for the source node of the edge \a edge
//	 * 		  (always less than \a trgt)
//	 * @param trgt the internal local id for the target node of the edge \a edge
//	 * 		  (always greater than \a src)
//	 * @param edge the edge with global node identifiers to map to a local edge
//	 */
//	void get_internal_ids(idx_t *src, idx_t *trgt, const edge_t &edge) const;

	// static members/methods
	static RELTYPE fcomp[MAXSET][MAXSET];	/**< composition table */
	static RELTYPE invtab[MAXSET];			/**< inverse relations */

	/**
	 * Calculates the size of the triangulated matrix representation of a
	 * network of \a sz number of nodes.
	 *
	 * For a network of N nodes, the triangulated matrix representation
	 * requires the following number of cells:
	 *								 			(N*N + N) / 2
	 * FIXME Why not (N*N - N) / 2 (minus N for excluding the diagonal)
	 *
	 * @param sz the number of nodes of the network
	 */
	inline static long matrix_size(int sz) {
		return ((long) sz * (long) sz + (long) sz) / (long) 2;
	}

	/**
	 * Compute the inverse of a RCC-8 relation.
	 *
	 * @param r an RCC-8 relation
	 * @return the inverse RCC-8 relation of \a r
	 */
	static RELTYPE inverse(RELTYPE r);

	/**
	 * Compute the relation value for a RCC-8 relation.
	 *
	 * @param r an RCC-8 relation
	 * @return the relation value of \a r
	 */
	static int relvalue(RELTYPE r);

	/**
	 * Computes the composition table of all possible disjunctions of
	 * the base RCC-8 relations. The table is stored in \a fcomp
	 */
	static void compute_comp_table();

	static int parsecons(const char *line, RELTYPE *r, int *node1, int *node2);
	static int parserel(const char *line, RELTYPE *r);


public:
	static std::string print_rel(RELTYPE r);

	/**
	 * Creates an empty (uninitialized) RCC-8 constraint network.
	 */
	RCC8_CSP();

	/**
	 * Creates a RCC-8 constraint network with the specified
	 * number of nodes and no edges.
	 *
	 * @param nodes the number of nodes of the network to create
	 */
	RCC8_CSP(int nodes);

	/**
	 * Creates an RCC-8 constraint network with the specified number
	 * of nodes, initialized to the given set of edges and using the
	 * specified mapping of node identifiers to internal identifiers.
	 *
	 * @param nodes the number of nodes
	 * @param matrix a const pointer to the native matrix holding the
	 *        RCC-8 relations
	 * @param fixed_matrix
	 * @param mapping a const reference to the mapping of node
	 *        identifiers to internal node identifiers
	 */
	RCC8_CSP(int nodes, const RELTYPE *matrix, const char *fixed_matrix, const nodeid_map &mapping);

	/**
	 *
	 * @param node_ids
	 * @param edges
	 * @param adjwgt
	 */
	/**
	 * Creates a RCC-8 constraint network by given a set of node ids,
	 * a set of edges (unlabeled), and the representation of the underlying
	 * graph in the CSR format of MeTiS.
	 *
	 * @param node_ids the set of node ids
	 * @param edges the set of unlabeled edges between the node ids
	 * @param xadj the matrix of node ids that indexes the matrix \a adjwgt
	 * @param adjncy the matrix of neighboring nodes indexed by \a xadj
	 * @param adjwgt the matrix holding the RCC-8 relations (labels) of the edges in \a adjncy
	 */
	RCC8_CSP(const std::set<idx_t> &node_ids,
			 const edge_set &edges,
			 const idx_t *xadj, const idx_t *adjncy, const RELTYPE *adjwgt);

	/**
	 * Creates an RCC-8 constraint network given in the CSR representation
	 * of MeTiS input sparse graph format.
	 *
	 * @param nodeids the subset of nodes that the network should contain
	 * @param xadj the xadj matrix of MeTiS CSR graph format
	 * @param adjncy the adjncy matrix of MeTiS CSR graph format
	 * @param adjwgt the adjwgt matrix of MeTiS CSR graph format
	 */
	//RCC8_CSP(std::vector<idx_t> nodeids, const idx_t *xadj, const idx_t *adjncy, const RELTYPE *adjwgt);

	/**
	 * Initializes the composition table for basic RCC-8 relations.
	 */
	static void initfop();

	/**
	 * Runs the well-known path-consistency algorithm on the underlying
	 * RCC-8 network.
	 *
	 * This is a modified version of the path_cons1() routine distributed
	 * with the RCC-8 reasoner of Ronny Fehling, Bernhard Nebel, and
	 * Jochen Renz.
	 *
	 * @param enable_fixed
	 * @return true if the network is path-consistent, false otherwise.
	 *///bool enable_fixed = false
	bool path_consistent();

	/**
	 * Checks whether the underlying RCC-8 network is consistent.
	 *
	 * This is an iterative algorithm that solves the consistency problem and not a
	 * backtracking one.
	 *
	 * By default this method runs the path consistency algorithm as a pre-processing
	 * step. One can disable this by passing true to parameter \a disable_pc.
	 *
	 * NOTICE: Change parameter \a disable_pc to true only if you are certain that the
	 *         network is path consistent. If it is not path consistent, then false
	 *         positive answers might be returned, since this method assumes that the
	 *         network is path consistent when it contains only RCC-8 relations from a
	 *         tractable subset.
	 *
	 * @param disable_pc [optional] controls whethe the path-consistency algorithm will
	 *                   be run as a pre-processing step
	 * @return true when the underlying network is consistent, false otherwise
	 *///bool disable_pc = false
	bool consistent();

	/**
	 * Creates a split_set data structure by iterating over the edges of the network and
	 * checking whether an RCC-8 relation can be split in horn relations. The data structure
	 * then contains all edges that can be split in horn relations. If such an edge does not
	 * exist, then it returns NULL.
	 *
	 * @return
	 */
	split_set* create_split_set();

	bool is_horn();

	/**
	 * Computes the split set in horn relations for the underlying network.
	 *
	 * @return true if the network has a split set, false otherwise
	 */
	inline bool compute_split_set() {
		split_it = create_split_set();
		return split_it != NULL;
	}

	inline const RELTYPE *get_csp_network() {
		return csp;
	}

	inline void set_csp_network(const RELTYPE *net) {
		memcpy(csp, net, RCC8_CSP::matrix_size(nnodes) * sizeof(RELTYPE));
	}

	void set_fixed_edges(const edge_list &fixed_edges) {
		edge_list::const_iterator it;
		for (it = fixed_edges.begin(); it != fixed_edges.end(); ++it) {
			idx_t src = nodeids.at(it->first);
			idx_t trgt = nodeids.at(it->second);

			ENSURE_ORDER(src, trgt);

			fixed[TRINDX(src, trgt)] = 1;
		}
	}

	/**
	 * Chooses the next relation to split and returns a refinement iterator
	 * for that relation.
	 */
	split_set *compute_next_split() const;

	/**
	 * Modifies the underlying RCC-8 network by applying the splits of RCC-8 relations that are
	 * represented by a specific split_set.
	 *
	 * @param the split_set representing the split of RCC-8 relations of the network
	 */
	void apply_local_split(split_set *split);

	/**
	 * Resets the underlying split set iterator.
	 *
	 * @return true if the iterator was reset, false if there was no iterator to reset
	 */
	bool reset_split();

	inline const split_set *get_split() const {
		return split_it;
	}

	inline RCC8_CSP* get_horn_csp() {
		return horn_csp;
	}

	bool apply_global_split(split_set *split);

	bool apply_global_split(bsplit_set *split);

	RELTYPE get_relation(idx_t real_srcid, idx_t real_trgtid) {
		idx_t src = nodeids.at(real_srcid);
		idx_t trgt = nodeids.at(real_trgtid);

		ENSURE_ORDER(src, trgt);

		return csp[TRINDX(src, trgt)];
	}

	/**
	 * Get the size of the RCC-8 constraint network (i.e., the number of nodes).
	 *
	 * @return the size
	 */
	inline int size() const {
		return nnodes;
	}

	/**
	 * Prints the upper triangular matrix corresponding to the underlying
	 * RCC-8 constraint network.
	 *
	 * NOTICE: Should be called only for relative small sizes of networks.
	 */
	inline void print() const {
		for (int i = 0; i < nnodes; i++) {
			for (int j = 0; j < nnodes; j++) {
				if (j >= i + 1) {
					printf("%3d ", csp[TRINDX(i, j)]);

				} else if (i == j) {
					printf("%3d ", 128);

				} else {
					printf("    ");
				}
			}
			std::cerr << "\n";
		}
	}

	/**
	 * Read a CSP file (extension .csp) and construct the underlying
	 * network.
	 *
	 * @param filename
	 * @return
	 */
	//static bool read(const char* filename);

	/**
	 * Calculates the memory requirements (in bytes) for this class
	 * for an RCC-8 network of \a sz number of nodes.
	 *
	 * @param sz the number of nodes of the network
	 */
	static long memory_needs(int sz) {
		return (2 * sizeof(char) + RELTYPESIZE) * matrix_size(sz);
	}

	/**
	 * Read a CSP file (extension .csp) and construct the underlying network
	 * in CRS format.
	 *
	 * FIXME: this method is incomplete!
	 */
	static bool read(const char* filename, idx_t *sz, idx_t **xadj, idx_t **ajncy, RELTYPE **adjwgt);

	/**
	 * Given two RCC-8 networks and a set of edges between nodes that have in common, check whether
	 * the respective labels on the edges agree and if not refine them in order to agree.
	 *
	 * TODO: in case this network is a horn one, you can optimize it further by eliminating also
	 * all EQ relations from any edge (see [Gerevini-Renz, CP'98])
	 *
	 * @param cspl a RCC-8 network
	 * @param cspr a RCC-8 network
	 * @param shared_edges common edges between the RCC-8 networks \a cspl and \a cspr
	 * @return -1: in case we cannot refine them (inconsistency),
	 * 		    0: the two networks agree on their common edges,
	 * 		    1: in case the refinement changed network \a cspr,
	 * 		    2: in case the refinement changed network \a cspl,
	 * 		    3: in case the refinement changed both of the networks
	 */
	static int check_and_refine(RCC8_CSP *cspl, RCC8_CSP *cspr, const edge_list &shared_edges);

	inline RCC8_CSP *clone() {
		return new RCC8_CSP(nnodes, csp, fixed, nodeids);
	}

	inline ~RCC8_CSP() {
		FREE_ARRAY(csp);
		FREE_ARRAY(mark);
		FREE_ARRAY(fixed);
		FREE_SPLIT(split_it);
		FREE_PTR(horn_csp);
	}
};

#endif /* RCC8CSP_H_ */
